# Copyright 2018 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.

from ctypes import (
    LittleEndianStructure,
    c_uint32,
    c_float
)

from .dataCommon import readCstr


class AMB_HEADER(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("magic",    c_uint32),
        ("version",  c_uint32),
        ("nbtracks", c_uint32)
    ]

class AMB_TRACK_HEADER(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("iflags", c_uint32),
        ("key_c",  c_uint32)
    ]

class AMB_TRACK_KeySetting(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("min",      c_float),
        ("max",      c_float),
        ("interval", c_uint32),
        ("flags",    c_uint32)
    ]

class AMB_TRACK_Key(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("flags",     c_uint32),
        ("start",     c_uint32),
        ("loop",      c_uint32),
        ("delay_min", c_uint32),
        ("delay_max", c_uint32),
        ("volume",    AMB_TRACK_KeySetting),
        ("pitch",     AMB_TRACK_KeySetting),
        ("pan",       AMB_TRACK_KeySetting),
        ("x",         AMB_TRACK_KeySetting),
        ("y",         AMB_TRACK_KeySetting),
        ("z",         AMB_TRACK_KeySetting)
    ]


import logging
from ctypes import sizeof
from collections import namedtuple
from typing import List

AmbTrack = namedtuple("AmbTrack", ["samplePath", "flags", "keys"])

class AmbSerializer(object):
    def __init__(self):
        self.log = logging.getLogger('AmbSerializer')

    def read(self, fileName) -> List[AmbTrack]:
        f = open(fileName, "rb")
        data = f.read()
        f.close()
        self.log.debug("Read %i bytes from file %s" % (len(data), fileName))

        pos = 0

        header = AMB_HEADER.from_buffer_copy(data, pos)
        pos += sizeof(AMB_HEADER)

        AMBIANCE_FILE_SIGNATURE = int('0x424d4147', 16)  # 'GAMB'
        VERSION_1001 = int('0x01000001', 16)
        VERSION_1002 = int('0x01000002', 16)
        VERSION_1003 = int('0x01000003', 16)

        assert(header.magic == AMBIANCE_FILE_SIGNATURE), "Magic mismatch"
        assert(header.version == VERSION_1001 or header.version == VERSION_1002 or header.version == VERSION_1003), "Bad version"

        tracks = []
        for i in range(header.nbtracks):
            samplePath, pos = readCstr(data, pos)
            samplePath = samplePath.replace('\\', '/').lower()

            if header.version >= VERSION_1002:
                trackName, pos = readCstr(data, pos)
                assert(not trackName)

            trackHeader = AMB_TRACK_HEADER.from_buffer_copy(data, pos)
            pos += sizeof(AMB_TRACK_HEADER)

            trackKeys = []
            for u in range(trackHeader.key_c):
                trackKey = AMB_TRACK_Key.from_buffer_copy(data, pos)
                pos += sizeof(AMB_TRACK_Key)
                trackKeys.append(trackKey)

            tracks.append(AmbTrack(
                samplePath=samplePath,
                flags=trackHeader.iflags,
                keys=trackKeys
            ))

        return tracks




# Copyright 2014-2019 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.

from ctypes import (
    LittleEndianStructure,
    c_char,
    c_int32,
    c_uint32,
    c_float
)

from .dataCommon import (
    SavedVec3,
    ArxQuat,
    SerializationException,
    UnexpectedValueException
)

class THEA_HEADER(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("identity",      c_char * 20), # THEO_SIZE_IDENTITY_ANIM
        ("version",       c_uint32),
        ("anim_name",     c_char * 256), # SIZE_NAME
        ("nb_frames",     c_int32),
        ("nb_groups",     c_int32),
        ("nb_key_frames", c_int32),
    ]

class THEA_KEYFRAME_2014(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("num_frame",        c_int32),
        ("flag_frame",       c_int32),
        ("master_key_frame", c_int32), # bool
        ("key_frame",        c_int32), # bool
        ("key_move",         c_int32), # bool
        ("key_orient",       c_int32), # bool
        ("key_morph",        c_int32), # bool
        ("time_frame",       c_int32),
    ]


class THEA_KEYFRAME_2015(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("num_frame",        c_int32),
        ("flag_frame",       c_int32),
        ("info_frame",       c_char * 256), # SIZE_NAME
        ("master_key_frame", c_int32), # bool
        ("key_frame",        c_int32), # bool
        ("key_move",         c_int32), # bool
        ("key_orient",       c_int32), # bool
        ("key_morph",        c_int32), # bool
        ("time_frame",       c_int32),
    ]

class THEA_MORPH(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("unknown1", c_int32),
        ("unknown2", c_int32),
        ("unknown3", c_int32),
        ("unknown4", c_int32),
    ]

class THEO_GROUPANIM(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("key_group",  c_int32),
        ("angle",      c_char * 8), # ignored
        ("Quaternion", ArxQuat),
        ("translate",  SavedVec3),
        ("zoom",       SavedVec3),
    ]

class THEA_SAMPLE(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("sample_name",  c_char * 256), # SIZE_NAME
        ("sample_size",  c_int32)
    ]


import logging
from ctypes import sizeof


from typing import List
from collections import namedtuple

TeaFrame = namedtuple("TeaFrame", ['duration', 'flags', 'translation', 'rotation', 'groups', 'sampleName'])

class TeaSerializer(object):
    def __init__(self):
        self.log = logging.getLogger('TeaSerializer')

    def read(self, fileName) -> List[TeaFrame]:
        f = open(fileName, "rb")
        data = f.read()
        f.close()
        self.log.debug("Read %i bytes from file %s" % (len(data), fileName))

        pos = 0

        header = THEA_HEADER.from_buffer_copy(data, pos)
        pos += sizeof(THEA_HEADER)

        self.log.debug(
            "Header - Identity: {0}; Version: {1}; Frames: {2}; Groups {3}; KeyFrames {4}".format(header.identity,
                                                                                                  header.version,
                                                                                                  header.nb_frames,
                                                                                                  header.nb_groups,
                                                                                                  header.nb_key_frames))

        if header.nb_frames < 0:
            raise UnexpectedValueException("header.nb_frames = " + str(header.nb_frames))

        if header.nb_groups < 0:
            raise UnexpectedValueException("header.nb_groups = " + str(header.nb_groups))

        if header.nb_key_frames < 0:
            raise UnexpectedValueException("header.nb_key_frames = " + str(header.nb_key_frames))

        results = []
        for i in range(header.nb_key_frames):
            # self.log.debug("Reading Keyframe {0}".format(i))
            if header.version == 2014:
                kf = THEA_KEYFRAME_2014.from_buffer_copy(data, pos)
                pos += sizeof(THEA_KEYFRAME_2014)
            elif header.version == 2015:
                kf = THEA_KEYFRAME_2015.from_buffer_copy(data, pos)
                pos += sizeof(THEA_KEYFRAME_2015)

                if kf.info_frame:
                    self.log.info("Keyframe str: " + kf.info_frame.decode('iso-8859-1'))
            else:
                raise SerializationException("Unknown version: " + str(header.version))

            if kf.flag_frame not in (-1, 9):
                raise UnexpectedValueException("flag_frame = " + str(kf.flag_frame))

            if kf.master_key_frame not in (0, 1):
                raise UnexpectedValueException("master_key_frame = " + str(kf.master_key_frame))
            if kf.key_frame not in (0, 1):
                raise UnexpectedValueException("key_frame = " + str(kf.key_frame))
            if kf.key_move not in (0, 1):
                raise UnexpectedValueException("key_move = " + str(kf.key_move))
            if kf.key_orient not in (0, 1):
                raise UnexpectedValueException("key_orient = " + str(kf.key_orient))
            if kf.key_morph not in (0, 1):
                raise UnexpectedValueException("key_morph = " + str(kf.key_morph))

            duration = kf.num_frame
            flags = kf.flag_frame

            # Global translation
            translation = None
            if kf.key_move != 0:
                translation = SavedVec3.from_buffer_copy(data, pos)
                pos += sizeof(SavedVec3)

            # Global rotation
            rotation = None
            if kf.key_orient != 0:
                pos += 8;  # skip THEO_ANGLE
                rotation = ArxQuat.from_buffer_copy(data, pos)
                pos += sizeof(ArxQuat)

            # Global Morph
            if kf.key_morph != 0:
                morph = THEA_MORPH.from_buffer_copy(data, pos)
                pos += sizeof(THEA_MORPH)

                if morph.unknown1 != -1:
                    raise UnexpectedValueException("unknown1 = " + str(morph.unknown1))
                if morph.unknown2 != -1:
                    raise UnexpectedValueException("unknown2 = " + str(morph.unknown2))

                self.log.debug("Frame {0} Morph: {1} {2}".format(i, morph.unknown3, morph.unknown4))
                # pos += 16; # skip THEA_MORPH

            # Now go for Group Rotations/Translations/scaling for each GROUP
            groupsList = THEO_GROUPANIM * header.nb_groups
            groups = groupsList.from_buffer_copy(data, pos)
            pos += sizeof(groupsList)

            # Has a sound sample ?
            num_sample = c_int32.from_buffer_copy(data, pos)
            pos += sizeof(c_int32)

            sampleName = None
            if num_sample.value != -1:
                sample = THEA_SAMPLE.from_buffer_copy(data, pos)
                pos += sizeof(THEA_SAMPLE)

                sampleName = sample.sample_name.decode('iso-8859-1');

                self.log.debug("sample_size: {0}".format(sample.sample_size))
                pos += sample.sample_size  # skip data

            pos += 4  # skip num_sfx
            # self.log.debug("Pos: {0}".format(pos))
            results.append(TeaFrame(
                duration=duration,
                flags=flags,
                translation=translation,
                rotation=rotation,
                groups=groups,
                sampleName=sampleName
            ))

        # Sanity check the deserialized data
        first = True
        grouplen = 0
        for f in results:
            if first:
                grouplen = len(f.groups)
                first = False
            else:
                if grouplen != len(f.groups):
                    raise UnexpectedValueException("Group count does not match!")

            for g in f.groups:
                if g.key_group not in (0, 1):
                    raise UnexpectedValueException("key_group = {}".format(g.key_group))

        self.log.debug("File loaded")

        return results



# Copyright 2019-2020 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.

import logging

import bpy
from mathutils import Vector, Quaternion

from .dataTea import TeaSerializer


class ArxAnimationManager(object):
    def __init__(self):
        self.log = logging.getLogger('ArxAnimationManager')
        self.teaSerializer = TeaSerializer()
        
    def loadAnimation(self, path):
        data = self.teaSerializer.read(path)
        
        selectedObj = bpy.context.active_object

        #Walk up object tree to amature
        walkObj = selectedObj
        while not walkObj.name.endswith('-amt') and walkObj.parent:
            walkObj = walkObj.parent

        if not walkObj.name.endswith('-amt'):
            self.log.warning("Amature object nof found for: {}".format(obj.name))
            return

        armatureObj = walkObj
        obj = walkObj.children[0]

        bones = armatureObj.pose.bones
        
        bpy.context.scene.frame_set(1)
        for frame in data:
            bpy.context.scene.objects.active = obj
            
            if frame.translation:
                translation = frame.translation
                obj.location = (translation.x,translation.y,translation.z)
                obj.keyframe_insert(data_path='location')
                
            if frame.rotation:
                rotation = frame.rotation
                obj.rotation_quaternion = (rotation.w,rotation.x,rotation.y,rotation.z)
                obj.keyframe_insert(data_path='rotation_quaternion')

            #bpy.ops.anim.keyframe_insert(type='LocRotScale', confirm_success=False)
            
            bpy.context.scene.objects.active = armatureObj
            bpy.ops.object.mode_set(mode='POSE')

            if len(bones) != len(frame.groups):
                #raise InconsistentStateException("Bones in amature must match animation groups, existing {} new {}".format(len(bones), len(frame['groups'])))
                self.log.warning("Bones in amature must match animation groups, existing {} new {}".format(len(bones), len(frame.groups)))
                #break

            # This seems to be required to handle mismatched data
            maxBone = min(len(bones), len(frame.groups))

            for groupIndex in range(maxBone - 1, -1, -1): # group index = bone index
                group = frame.groups[groupIndex]
                bone = bones[groupIndex]
                location = Vector((group.translate.x,group.translate.y,group.translate.z))
                #self.log.info("moving bone to %s" % str(group.translate))
                #bone.location = location
                rotation = Quaternion((group.Quaternion.w,group.Quaternion.x,group.Quaternion.y,group.Quaternion.z))
                #self.log.info("rotating bone to %s" % str(rotation))
                bone.rotation_quaternion = rotation
                scale = Vector((group.zoom.x,group.zoom.y,group.zoom.z))
                #self.log.info("scaling bone to %s" % str(group.zoom))
                #bone.scale = scale
                bone.keyframe_insert(data_path="location")
                bone.keyframe_insert(data_path="rotation_quaternion")
                bone.keyframe_insert(data_path="scale")
                #bpy.ops.anim.keyframe_insert(type='LocRotScale', confirm_success=False)
                # FIXME translation and zoom are both 0,0,0
                # TODO keyframes of rotation are glitchy. probably cause im not doing it right. keyframe has to be set via bpy.ops.anim.keyframe_insert but i dont know how to select the bone.
                
            bpy.ops.object.mode_set(mode='OBJECT')
            
            bpy.context.scene.frame_set(bpy.context.scene.frame_current + frame.duration)
            #self.log.info("Loaded Frame")
        bpy.context.scene.frame_end = bpy.context.scene.frame_current


# Copyright 2014-2019 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.

from ctypes import (
    LittleEndianStructure,
    Union,
    c_uint8,
    c_uint32,
    c_float
)

import struct


class SerializationException(Exception):
    pass


class UnexpectedValueException(SerializationException):
    pass


def readCstr(data, pos):
    strEnd = data[pos:].index(b'\x00')
    strEndPos = pos + strEnd
    somStr = data[pos:strEndPos]
    decoded = somStr.decode('iso-8859-1')
    return (decoded, strEndPos + 1) # +1 for the null


def read_s32(data, pos):
    value = struct.unpack_from('<i', data, pos)[0]
    endPos = pos + 4;
    return (value, endPos)


class SavedVec3(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("x", c_float),
        ("y", c_float),
        ("z", c_float)
    ]
    
    def __str__(self):
        return "<SavedVec3 (x={self.x}, y={self.y}, z={self.z})>".format(self=self)


class SavedAnglef(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("a", c_float),
        ("b", c_float),
        ("g", c_float)
    ]
    
    def __str__(self):
        return "<SavedAnglef (a={self.a}, b={self.b}, g={self.g})>".format(self=self)


class ArxQuat(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("w", c_float),
        ("x", c_float),
        ("y", c_float),
        ("z", c_float)
    ]
    
    def __str__(self):
        return "<ArxQuat (w={self.w}, x={self.x}, y={self.y}, z={self.z})>".format(self=self)


class SavedColor(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("r", c_float),
        ("g", c_float),
        ("b", c_float)
    ]
    
    def __str__(self):
        return "<SavedColor (r={self.r}, g={self.g}, b={self.b})>".format(self=self)


class PolyTypeFlag_bits(LittleEndianStructure):
     _fields_ = [
        ("POLY_NO_SHADOW",    c_uint8, 1),
        ("POLY_DOUBLESIDED",  c_uint8, 1),
        ("POLY_TRANS",        c_uint8, 1),
        ("POLY_WATER",        c_uint8, 1),
        ("POLY_GLOW",         c_uint8, 1),
        ("POLY_IGNORE",       c_uint8, 1),
        ("POLY_QUAD",         c_uint8, 1),
        ("POLY_TILED",        c_uint8, 1), # Unused
        ("POLY_METAL",        c_uint8, 1),
        ("POLY_HIDE",         c_uint8, 1),
        ("POLY_STONE",        c_uint8, 1),
        ("POLY_WOOD",         c_uint8, 1),
        ("POLY_GRAVEL",       c_uint8, 1),
        ("POLY_EARTH",        c_uint8, 1),
        ("POLY_NOCOL",        c_uint8, 1),
        ("POLY_LAVA",         c_uint8, 1),
        ("POLY_CLIMB",        c_uint8, 1),
        ("POLY_FALL",         c_uint8, 1),
        ("POLY_NOPATH",       c_uint8, 1),
        ("POLY_NODRAW",       c_uint8, 1),
        ("POLY_PRECISE_PATH", c_uint8, 1),
        ("POLY_NO_CLIMB",     c_uint8, 1), # Unused
        ("POLY_ANGULAR",      c_uint8, 1), # Unused
        ("POLY_ANGULAR_IDX0", c_uint8, 1), # Unused
        ("POLY_ANGULAR_IDX1", c_uint8, 1), # Unused
        ("POLY_ANGULAR_IDX2", c_uint8, 1), # Unused
        ("POLY_ANGULAR_IDX3", c_uint8, 1), # Unused
        ("POLY_LATE_MIP",     c_uint8, 1),
    ]


class PolyTypeFlag(Union):
    _fields_ = [
        ("b",      PolyTypeFlag_bits),
        ("asUInt", c_uint32),
    ]
    _anonymous_ = ("b")




import dataclasses
import logging
import os

import bpy

from .arx_io_util import InconsistentStateException
from .dataFtl import FtlFace
from .files import ArxFiles

logging.basicConfig(level=logging.INFO)

# ======================================================================================================================

class DiffNode(object):
    def __init__(self, name, parent=None):
        self.name = name
        self.parent = parent
        self.children = []
        if parent:
            parent.children.append(self)

    def addChild(self, child):
        child.parent = self
        self.children.append(child)

    def __str__(self):
        result = ""

        stack = list()
        stack.append(self)

        child_stack_list = list()
        child_stack_list.append(stack)

        while child_stack_list:
            child_stack = child_stack_list[-1]
            if child_stack:
                current_node = child_stack.pop(0)

                stack_indent = "";
                for i in range(0, len(child_stack_list) - 1):
                    if child_stack_list[i]:
                        stack_indent += "│ "
                    else:
                        stack_indent += "  "

                if child_stack:
                    current_indent = '├ '
                else:
                    current_indent = '╰ '

                result += stack_indent + current_indent + current_node.name + '\n'

                if current_node.children:
                    # Make sure to copy the list here
                    child_stack_list.append(list(current_node.children))
            else:
                child_stack_list.pop()

        return result

# ======================================================================================================================

def equal_float(a, b, tol = 0.00001):
    return abs(a - b) < tol

def value_equal(expected, actual):
    if isinstance(expected, tuple):
        for expected_e, actual_e in zip(expected, actual):
            if type(expected_e) != type(actual_e):
                return False
            if type(expected_e) == float:
                if not equal_float(expected_e, actual_e):
                    return False
            else:
                if expected_e != actual_e:
                    return False
        return True
    else:
        return expected == actual


def arx_deep_compare(a, b, parent) -> DiffNode:
    name = type(a).__name__
    node = DiffNode(name, parent)
    if value_equal(a, b):
        DiffNode('OK', node)
    else:
        if dataclasses.is_dataclass(a):
            for field in dataclasses.fields(a):
                field_a = getattr(a, field.name)
                field_b = getattr(b, field.name)
                field_node = DiffNode(field.name, node)
                if value_equal(field_a, field_b):
                    DiffNode('OK', field_node)
                    continue
                arx_deep_compare(field_a, field_b, field_node)
        elif isinstance(a, list):
            if len(a) != len(b):
                DiffNode("length; expecteded {0} actual {1}".format(len(a), len(b)), node)
            limit = 0
            for i, (field_a, field_b) in enumerate(zip(a, b)):
                if value_equal(field_a, field_b):
                    continue
                index_node = DiffNode("Index: {0}".format(i), node)
                arx_deep_compare(field_a, field_b, index_node)
                limit += 1
                if limit > 2:
                    DiffNode("Omitted more errors ...", index_node)
                    break
        else:
            DiffNode("expecteded {0}".format(a), node)
            DiffNode("actual     {0}".format(b), node)

def arx_ftl_compare(a, b) -> DiffNode:
    root = DiffNode('root')
    arx_deep_compare(a, b, root)
    return root

# ======================================================================================================================

def formatColumns(values):
    columnWidth = 15
    return '\t'.join([val.ljust(columnWidth) for val in values]) + '\n'

def test_export_of_current_scene_objects(assetManager):
    result_summary = bpy.data.texts.new('arx_test_export')
    result_summary.write(formatColumns(["Compare", "File"]))

    for bobj in bpy.data.objects:
        aobjId = tuple(bobj.name.split("/"))
        #TODO this is a hack to check if its a real object
        if len(aobjId) < 2:
            continue

        fsInfo = assetManager.arxFiles.models.data[aobjId]
        fileName = os.path.join(fsInfo.path, fsInfo.model)
        dataA = assetManager.objectManager.loadFile_data(fileName)

        dataB = assetManager.objectManager.toFtlData(bobj)

        comparison = arx_ftl_compare(dataA, dataB)

        txt = bpy.data.texts.new('arx_test_export_' + bobj.name)
        if not comparison.children:
            result_summary.write(formatColumns(['ok', bobj.name]))
            txt.from_string("Ok")
        else:
            txt.from_string(str(comparison))
            result_summary.write(formatColumns(['ERROR', bobj.name]))



# Copyright 2014-2020 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.

from ctypes import (
    LittleEndianStructure,
    c_char,
    c_uint32,
    c_int16,
    c_int32,
    c_float
)

from .dataCommon import (
    SavedVec3,
    SavedAnglef,
    PolyTypeFlag,
    SavedColor
)

class DANAE_LS_HEADER(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("version",         c_float),
        ("ident",           c_char * 16),
        ("lastuser",        c_char * 256),
        ("time",            c_int32),
        ("pos_edit",        SavedVec3),
        ("angle_edit",      SavedAnglef),
        ("nb_scn",          c_int32),
        ("nb_inter",        c_int32),
        ("nb_nodes",        c_int32),
        ("nb_nodeslinks",   c_int32),
        ("nb_zones",        c_int32),
        ("lighting",        c_int32),
        ("Bpad",            c_int32 * 256),
        ("nb_lights",       c_int32),
        ("nb_fogs",         c_int32),
        ("nb_bkgpolys",     c_int32),
        ("nb_ignoredpolys", c_int32),
        ("nb_childpolys",   c_int32),
        ("nb_paths",        c_int32),
        ("pad",             c_int32 * 250),
        ("offset",          SavedVec3),
        ("fpad",            c_float * 253),
        ("cpad",            c_char * 4096),
        ("bpad",            c_int32 * 256),
    ]

class DANAE_LS_SCENE(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("name", c_char * 512),
        ("pad",  c_int32 * 16),
        ("fpad", c_float * 16)
    ]

class DANAE_LS_INTER(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("name",     c_char * 512),
        ("pos",      SavedVec3),
        ("angle",    SavedAnglef),
        ("ident",    c_int32),
        ("flags",    c_int32),
        ("pad",      c_int32 * 14),
        ("fpad",     c_float * 16),
    ]

class DANAE_LS_LIGHTINGHEADER(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("nb_values", c_int32),
        ("ViewMode",  c_int32), # unused
        ("ModeLight", c_int32), # unused
        ("pad",       c_int32), # unused
    ]

# version 1.003f
# TODO probably unused
class DANAE_LS_LIGHT(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("pos",          SavedVec3),
        ("rgb",          SavedColor),
        ("fallstart",    c_float),
        ("fallend",      c_float),
        ("intensity",    c_float),
        ("i",            c_float),
        ("ex_flicker",   SavedColor),
        ("ex_radius",    c_float),
        ("ex_frequency", c_float),
        ("ex_size",      c_float),
        ("ex_speed",     c_float),
        ("ex_flaresize", c_float),
        ("fpadd",        c_float * 24),
        ("extras",       c_int32),
        ("lpadd",        c_int32 * 31)
    ]

class DANAE_LS_FOG(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("pos",         SavedVec3),
        ("rgb",         SavedColor),
        ("size",        c_float),
        ("special",     c_int32),
        ("scale",       c_float),
        ("move",        SavedVec3),
        ("angle",       SavedAnglef),
        ("speed",       c_float),
        ("rotatespeed", c_float),
        ("tolive",      c_int32),
        ("blend",       c_int32),
        ("frequency",   c_float),
        ("fpadd",       c_float * 32),
        ("lpadd",       c_int32 * 32),
        ("cpadd",       c_char * 256)
    ]

class DANAE_LS_PATH(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("name",        c_char * 64),
        ("idx",         c_int16),
        ("flags",       c_int16),
        ("initpos",     SavedVec3),
        ("pos",         SavedVec3),
        ("nb_pathways", c_int32),
        ("rgb",         SavedColor),
        ("farclip",     c_float),
        ("reverb",      c_float),
        ("amb_max_vol", c_float),
        ("fpadd",       c_float * 26),
        ("height",      c_int32),
        ("lpadd",       c_int32 * 31),
        ("ambiance",    c_char * 128),
        ("cpadd",       c_char * 128)
    ]

class DANAE_LS_PATHWAYS(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("rpos",  SavedVec3),
        ("flag",  c_int32),
        ("time",  c_uint32),
        ("fpadd", c_float * 2),
        ("lpadd", c_int32 * 2),
        ("cpadd", c_char * 32)
    ]


import logging
from ctypes import sizeof, create_string_buffer

from dataclasses import dataclass
from typing import List, Tuple

@dataclass
class DlfData:
    entities: List[DANAE_LS_INTER]
    fogs:     List[DANAE_LS_FOG]
    paths:    List[Tuple[DANAE_LS_PATH, List[DANAE_LS_PATHWAYS]]]

class DlfSerializer(object):
    def __init__(self, ioLib):
        self.log = logging.getLogger('DlfSerializer')
        self.ioLib = ioLib
    
    def read(self, data, lsHeader) -> DlfData:
        pos = 0
        result = {}
        
        scnHeader = DANAE_LS_SCENE.from_buffer_copy(data, pos)
        pos += sizeof(DANAE_LS_SCENE)
        self.log.debug("DANAE_LS_SCENE name: %s" % scnHeader.name.decode('iso-8859-1'))
        
        EntitiesType = DANAE_LS_INTER * lsHeader.nb_inter
        entities = EntitiesType.from_buffer_copy(data, pos)
        pos += sizeof(EntitiesType)

        if lsHeader.lighting != 0:
            lightingHeader = DANAE_LS_LIGHTINGHEADER.from_buffer_copy(data, pos)
            pos += sizeof(DANAE_LS_LIGHTINGHEADER)
            
            pos += lightingHeader.nb_values * 4 # Skip the data
        
        # Skip lights
        pos += sizeof(DANAE_LS_LIGHT) * lsHeader.nb_lights

        FogsType = DANAE_LS_FOG * lsHeader.nb_fogs
        fogs = FogsType.from_buffer_copy(data, pos)
        pos += sizeof(FogsType)

        # Skip nodes
        if lsHeader.version >= 1.001:
            pos += lsHeader.nb_nodes * (204 + lsHeader.nb_nodeslinks * 64)

        paths = []
        for i in range(lsHeader.nb_paths):
            path = DANAE_LS_PATH.from_buffer_copy(data, pos)
            pos += sizeof(DANAE_LS_PATH)

            segments = []
            for u in range(path.nb_pathways):
                segment = DANAE_LS_PATHWAYS.from_buffer_copy(data, pos)
                pos += sizeof(DANAE_LS_PATHWAYS)
                segments.append(segment)

            paths.append((path, segments))

        return DlfData(
            entities=entities,
            fogs=fogs,
            paths=paths
        )
        
    
    def readContainer(self, fileName) -> DlfData:
        f = open(fileName, "rb")
        data = f.read()
        f.close()

        self.log.debug("Loaded %i bytes from file %s" % (len(data), fileName))
        
        pos = 0
        lsHeader = DANAE_LS_HEADER.from_buffer_copy(data, pos)
        pos += sizeof(DANAE_LS_HEADER)
        self.log.debug("DANAE_LS_HEADER version: %f" % lsHeader.version)
        
        if lsHeader.ident.decode('iso-8859-1') != "DANAE_FILE":
            raise Exception('Invalid ident')
        
        uncompressed = self.ioLib.unpack(data[pos:])
        
        return self.read(uncompressed, lsHeader)




# Copyright 2014-2017 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.

import bmesh

import bpy

from .dataCommon import PolyTypeFlag

faceFlagNames = [
    ("POLY_NO_SHADOW", "NO_SHADOW"),
    ("POLY_DOUBLESIDED", "DOUBLESIDED"),
    ("POLY_TRANS", "TRANS"),
    ("POLY_WATER", "WATER"),
    ("POLY_GLOW", "GLOW"),
    ("POLY_IGNORE", "IGNORE"),
    ("POLY_QUAD", "QUAD"),
    ("POLY_METAL", "METAL"),
    ("POLY_HIDE", "HIDE"),
    ("POLY_STONE", "STONE"),
    ("POLY_WOOD", "WOOD"),
    ("POLY_GRAVEL", "GRAVEL"),
    ("POLY_EARTH", "EARTH"),
    ("POLY_NOCOL", "NOCOL"),
    ("POLY_LAVA", "LAVA"),
    ("POLY_CLIMB", "CLIMB"),
    ("POLY_FALL", "FALL"),
    ("POLY_NOPATH", "NOPATH"),
    ("POLY_NODRAW", "NODRAW"),
    ("POLY_PRECISE_PATH", "PRECISE_PATH"),
    ("POLY_LATE_MIP", "LATE_MIP")
]


class ArxMeshAddCustomProperties(bpy.types.Operator):
    bl_idname = "arx.mesh_add_custom_properties"
    bl_label = "Add custom properties to mesh"

    def execute(self, context):
        obj = bpy.context.object
        if obj is None:
            self.report({'ERROR'}, "No object selected")
            return {'CANCELLED'}
        if obj.type != 'MESH':
            self.report({'ERROR'}, "Selected object not a mesh")
            return {'CANCELLED'}
        
        if obj.data.is_editmode:
            bm = bmesh.from_edit_mesh(obj.data)
        else:
            bm = bmesh.new()
            bm.from_mesh(obj.data)
        
        if not bm.faces.layers.int.get('arx_facetype'):
            self.report({'INFO'}, "Adding missing " + 'arx_facetype' + " layer")
            bm.faces.layers.int.new('arx_facetype')
        
        if not bm.faces.layers.float.get('arx_transval'):
            self.report({'INFO'}, "Adding missing " + 'arx_transval' + " layer")
            bm.faces.layers.float.new('arx_transval')
        
        if bm.is_wrapped:
            bmesh.update_edit_mesh(obj.data, False, False)
        else:
            bm.to_mesh(obj.data)
            obj.data.update()
        
        return {'FINISHED'}

class ArxFacePanel(bpy.types.Panel):
    bl_label = "Arx Face Properties"
    bl_space_type = 'VIEW_3D'
    bl_region_type = "UI"

    bpy.types.Scene.ArxFaceFlags = bpy.props.BoolVectorProperty(name="ArxFaceFlags",
                                                                default=([False] * len(faceFlagNames)),
                                                                size=len(faceFlagNames))

    def draw(self, context):
        obj = bpy.context.object

        if obj is None:
            return
        
        layout = self.layout

        if obj.type != 'MESH':
            layout.label(text="Not a mesh")
            return

        if obj.data.is_editmode:
            bm = bmesh.from_edit_mesh(obj.data)
        else:
            bm = bmesh.new()
            bm.from_mesh(obj.data)

        arxFaceType = bm.faces.layers.int.get('arx_facetype')
        arxTransVal = bm.faces.layers.float.get('arx_transval')

        if arxFaceType is None or arxTransVal is None:
            layout.label(text="Not an arx mesh")
            layout.operator("arx.mesh_add_custom_properties");
            return

        face = bm.faces.active

        if face is None:
            return

        faceType = PolyTypeFlag()
        faceType.asUInt = face[arxFaceType]

        transval = face[arxTransVal]

        scene = context.scene

        layout.label(text="transval: " + str(transval))

        for i, (prop, text) in enumerate(faceFlagNames):
            a = getattr(faceType, prop)
            scene.ArxFaceFlags[i] = a

        row = layout.column(align=True)
        for i, (prop, text) in enumerate(faceFlagNames):
            row.prop(scene, "ArxFaceFlags", index=i, text=text, toggle=True)



# Copyright 2015-2020 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.


import importlib
import sys

if "ArxAddon" in locals():
    importlib.reload(sys.modules["arx_addon.managers"])

if "ArxFacePanel" in locals():
    importlib.reload(sys.modules["arx_addon.meshEdit"])

from .arx_io_util import ArxException
from .managers import ArxAddon
from .meshEdit import ArxFacePanel, ArxMeshAddCustomProperties

import bpy

from bpy_extras.io_utils import ExportHelper, ImportHelper, path_reference_mode

from bpy.types import AddonPreferences

from bpy.props import BoolProperty, StringProperty

import logging

from .arx_ui_area import (
    arx_ui_area_register,
    arx_ui_area_unregister
)

from .managers import (
    arxAddonReload,
    getAddon
)

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger('ArxAddon')

# ======================================================================================================================

class ArxAddonPreferences(bpy.types.AddonPreferences):
    bl_idname = __package__

    def reload(self, context):
        arxAddonReload()

    arxAssetPath: bpy.props.StringProperty(
        name="Arx assets root directory",
        subtype='DIR_PATH',
        update=reload
    )
    arxAllowLibFallback: bpy.props.BoolProperty(
        name="Allow use of fallback io library, only import ftl models, scenes are broken!",
        update=reload
    )

    def draw(self, context):
        layout = self.layout
        layout.prop(self, "arxAssetPath")
        layout.prop(self, "arxAllowLibFallback")

# ======================================================================================================================




"""
class ArxOperatorCreateLevelEditScreen(bpy.types.Operator):
    bl_idname = "arx.operator_create_level_edit_screen"
    bl_label = "Create level editing screen"
    
    def execute(self, context):
        name = "Arx Level Edit"
        
        if bpy.data.screens.get(name):
            self.report({'ERROR'}, "Screen already exists")
            return {'CANCELLED'}
        
        bpy.ops.screen.new()
        screen = bpy.context.window.screen
        screen.name = "Arx Level Edit"
        
        for a in bpy.context.screen.areas:
            if a.type == 'VIEW_3D':
                for s in a.spaces:
                    if s.type == 'VIEW_3D':
                        s.clip_end = 100000
                        s.show_relationship_lines = False
                        s.show_backface_culling = True
        
        return {'FINISHED'}
"""

# ======================================================================================================================

class ArxModelsPanel(bpy.types.Panel):
    bl_idname = "SCENE_PT_models_panel"
    bl_label = "Arx Libertatis Models"
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "scene"

    def draw(self, context):
        layout = self.layout
        layout.operator("arx.operator_import_all_models")
        layout.operator("arx.operator_test_model_export")

# ======================================================================================================================

class ArxOperatorTestModelExport(bpy.types.Operator):
    bl_idname = "arx.operator_test_model_export"
    bl_label = "Test model export"
    
    def execute(self, context):
        try:
            getAddon(context).testModelExport()
            return {'FINISHED'}
        except ArxException as e:
            self.report({'ERROR'}, str(e))
            return {'CANCELLED'}


class ArxOperatorImportAllModels(bpy.types.Operator):
    bl_idname = "arx.operator_import_all_models"
    bl_label = "Import all models"
    
    def execute(self, context):
        try:
            getAddon(context).assetManager.importAllModels(context)
            return {'FINISHED'}
        except ArxException as e:
            self.report({'ERROR'}, str(e))
            return {'CANCELLED'}



class ImportFTL(bpy.types.Operator, ImportHelper):
    '''Load an Arx Fatalis Model File'''
    bl_idname = "arx.import_ftl"
    bl_label = 'Import FTL model'
    bl_options = {'PRESET'}

    filename_ext = ".ftl"
    check_extension = True
    #path_mode = path_reference_mode

    filter_glob: StringProperty(
        default="*.ftl",
        options={'HIDDEN'}
    )

    import_tweaks: BoolProperty(
        name="Import Tweaks",
        default=False
    )

    def execute(self, context):
        try:
            scene = bpy.context.scene
            getAddon(context).objectManager.loadFile(context, self.filepath, scene, self.import_tweaks)
            return {'FINISHED'}
        except ArxException as e:
            self.report({'ERROR'}, str(e))
            return {'CANCELLED'}


def menu_func_import_ftl(self, context):
    self.layout.operator(ImportFTL.bl_idname, text="Arx Fatalis Model (.ftl)")


class ExportFTL(bpy.types.Operator, ExportHelper):
    '''Save a raw uncompressed Arx Fatalis Model File'''
    bl_idname = "arx.export_ftl"
    bl_label = 'Export FTL model'
    bl_options = {'PRESET'}

    filename_ext = ".ftl"
    check_extension = True
    #path_mode = path_reference_mode

    filter_glob: StringProperty(
        default="*.ftl",
        options={'HIDDEN'}
    )

    def execute(self, context):
        try:
            getAddon(context).objectManager.saveFile(self.filepath)
            return {'FINISHED'}
        except ArxException as e:
            self.report({'ERROR'}, str(e))
            return {'CANCELLED'}


def menu_func_export_ftl(self, context):
    self.layout.operator(ExportFTL.bl_idname, text="Arx Fatalis Model (.ftl)")


# ======================================================================================================================

class ImportTea(bpy.types.Operator, ImportHelper):
    """Load a Tea animation file"""
    bl_idname = "arx.import_tea"
    bl_label = "Import Tea animation"
    bl_options = {'REGISTER', 'UNDO'}

    filename_ext = ".tea"
    check_extension = True
    #path_mode = path_reference_mode

    filter_glob: StringProperty(
        default="*.tea",
        options={'HIDDEN'}
    )

    def execute(self, context):
        getAddon(context).animationManager.loadAnimation(self.filepath)
        return {'FINISHED'}
    

def menu_func_import_tea(self, context):
    self.layout.operator(ImportTea.bl_idname, text="Tea animation (.tea)")


# ============= Registration

classes = (
    ArxAddonPreferences,
#
    ArxOperatorImportAllModels,
    ArxOperatorTestModelExport,
    ArxModelsPanel,
#
    ArxMeshAddCustomProperties,
    ArxFacePanel,
    ImportFTL,
    ExportFTL,
    ImportTea
)

def register():
    log.debug("register")

    for c in classes:
        bpy.utils.register_class(c)

    arx_ui_area_register()

    bpy.types.TOPBAR_MT_file_import.append(menu_func_import_ftl)
    bpy.types.TOPBAR_MT_file_export.append(menu_func_export_ftl)
    bpy.types.TOPBAR_MT_file_import.append(menu_func_import_tea)


def unregister():
    log.debug("unregister")

    bpy.types.TOPBAR_MT_file_import.remove(menu_func_import_ftl)
    bpy.types.TOPBAR_MT_file_export.remove(menu_func_export_ftl)
    bpy.types.TOPBAR_MT_file_import.remove(menu_func_import_tea)

    for c in classes:
        bpy.utils.unregister_class(c)

    arx_ui_area_unregister()

    arxAddonReload()



# Copyright 2019-2020 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.

from mathutils import Vector

class ArxException(Exception):
    """Common exception thrown by this addon"""
    pass

class InconsistentStateException(Exception):
    """Thrown if data supposed to be added to existing data does not match"""
    pass

def arx_pos_to_blender_for_model(pos):
    """x=>x; y=>-z; z=>y"""
    return Vector((pos[0], pos[2], -pos[1]))

def blender_pos_to_arx(pos):
    return (pos[0], -pos[2], pos[1])



# Copyright 2014-2019 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.

from ctypes import (
    LittleEndianStructure,
    c_char,
    c_int16,
    c_int32,
    c_uint32,
    c_float
)

from .dataCommon import (
    SavedVec3,
    SavedColor,
    UnexpectedValueException,
    readCstr,
    read_s32
)

def structToString(struct):
    return ', '.join([f[0] + '=' + str(getattr(struct, f[0])) for f in struct._fields_])

class CIN_HEADER(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("magic",   c_char * 4),
        ("version", c_uint32)
    ]

class CIN_TRACK(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('startframe', c_int32), # unused always 0
        ('endframe',   c_int32),
        ('currframe',  c_float), # unused mostly 0
        ('fps',        c_float),
        ('nbkey',      c_int32),
        ('pause',      c_int32)  # unused always 1
    ]
    def __str__(self):
        return 'e {} curr {} fps {} keys {}'.format(
            self.endframe, self.currframe, self.fps, self.nbkey)


class CinematicLight_1_71(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('pos',          SavedVec3),
        ('fallin',       c_float),
        ('fallout',      c_float),
        ('color',        SavedColor),
        ('intensity',    c_float),
        ('intensiternd', c_float),
        ('prev',         c_int32), # unused
        ('next',         c_int32)  # unused
    ]
    def __str__(self):
        return structToString(self)

class CIN_KEY_1_75(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('frame',      c_int32),
        ('numbitmap',  c_int32),
        ('fx',         c_int32),
        ('typeinterp', c_int16),
        ('force',      c_int16),
        ('pos',        SavedVec3),
        ('angz',       c_float),
        ('color',      c_int32),
        ('colord',     c_int32),
        ('colorf',     c_int32),
        ('idsound',    c_int32),
        ('speed',      c_float),
        ('light',      CinematicLight_1_71),
        ('posgrille',  SavedVec3),
        ('angzgrille', c_float),
        ('speedtrack', c_float)
    ]
    def __str__(self):
        return structToString(self)

class Cin_Idsound_List(c_int32 * 16):
    def __str__(self):
        return ', '.join(str(self[i]) for i in range(len(self)))

class CIN_KEY_1_76(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('frame',      c_int32),
        ('numbitmap',  c_int32),
        ('fx',         c_int32), # associated fx
        ('typeinterp', c_int16),
        ('force',      c_int16),
        ('pos',        SavedVec3),
        ('angz',       c_float),
        ('color',      c_int32),
        ('colord',     c_int32),
        ('colorf',     c_int32),
        ('speed',      c_float),
        ('light',      CinematicLight_1_71),
        ('posgrille',  SavedVec3),
        ('angzgrille', c_float),
        ('speedtrack', c_float),
        ('idsound',    Cin_Idsound_List)
    ]
    def __str__(self):
        return structToString(self)


import logging
from ctypes import sizeof
import os

from collections import namedtuple

CinData = namedtuple('CinData',
    ['track', 'keyframes'])

CinKeyframe = namedtuple('CinKeyframe',
    ['frame', 'image', 'sound'])

def normalizeImagePath(path):
    parts = path.lower().split('\\')
    name, ext = os.path.splitext(parts[-1])
    return name

def normalizeSoundPath(path):
    parts = path.lower().split('\\')
    return parts[-1]

class CinSerializer(object):
    def __init__(self):
        self.log = logging.getLogger('CinSerializer')
        self.log.setLevel(logging.DEBUG)

    def read(self, fileName):
        f = open(fileName, "rb")
        data = f.read()
        f.close()
        self.log.debug("Read %i bytes from file %s" % (len(data), fileName))

        pos = 0

        header = CIN_HEADER.from_buffer_copy(data, pos)
        pos += sizeof(CIN_HEADER)

        assert(header.magic == b'KFA'), "Magic mismatch"
        assert(header.version == 65611 or header.version == 65612), "Version invalid"

        weirdStr, pos = readCstr(data, pos)
        #self.log.info('str = {}'.format(weirdStr))

        images = []
        nbitmaps, pos = read_s32(data, pos)
        for i in range(nbitmaps):
            scale, pos = read_s32(data, pos)
            texture, pos = readCstr(data, pos)
            normTexture = normalizeImagePath(texture)
            images.append((scale, normTexture))
            #self.log.info("Texture {} - Scale: {}; Path: {}".format(i, scale, texture))

        sounds = []
        nsounds, pos = read_s32(data, pos)
        for i in range(nsounds):
            if(header.version == 65612):
                pos += 2 # ignore sound id
            sound, pos = readCstr(data, pos)
            sounds.append(sound)
            #self.log.info("Sound {} - Path: {}".format(i, sound))

        track = CIN_TRACK.from_buffer_copy(data, pos)
        pos += sizeof(CIN_TRACK)

        #self.log.info('Cinematic {} - Version {} Images {} Sounds {} Keyframes {}'.format(
        #    fileName, header.version, nbitmaps, nsounds, track.nbkey))

        #self.log.info(
        #    "Track - {}".format(str(track)))

        if track.startframe != 0:
            raise UnexpectedValueException("startframe = " + str(track.startframe))
        if track.pause != 1:
            raise UnexpectedValueException("track.pause = " + str(track.pause))

        def resolve_sound_id(id):
            if id == -1:
                return None
            elif id < len(sounds):
                return normalizeSoundPath(sounds[id])
            else:
                self.log.warning('Invalid sound index {} id {} total sounds {}'.format(i, id, len(sounds)))
                return None


        initialSound = None

        keyframes = []
        for i in range(track.nbkey):
            if (header.version == 65611):
                keyframe = CIN_KEY_1_75.from_buffer_copy(data, pos)
                pos += sizeof(CIN_KEY_1_75)

                #self.log.debug('Raw Keyframe - {}'.format(keyframe))

                if i == 0:
                    initialSound = resolve_sound_id(keyframe.idsound)

                sound = None
            else:
                keyframe = CIN_KEY_1_76.from_buffer_copy(data, pos)
                pos += sizeof(CIN_KEY_1_76)

                #self.log.debug('Raw Keyframe - {}'.format(keyframe))

                if i == 0:
                    initialSound = resolve_sound_id(keyframe.idsound[0])

                sound = resolve_sound_id(keyframe.idsound[3])

            image = images[keyframe.numbitmap]

            foo = CinKeyframe(
                frame = keyframe.frame,
                image = image,
                sound = sound
            )

            #self.log.debug('Keyframe - {}'.format(foo))

            keyframes.append(foo)

            if keyframe.typeinterp not in (-1, 0, 1):
                self.log.error('keyframe.typeinterp = {}'.format(str(keyframe.typeinterp)))
                #raise UnexpectedValueException("keyframe.typeinterp = " + str(keyframe.typeinterp))

            if keyframe.posgrille.x != 0 or keyframe.posgrille.y != 0 or keyframe.posgrille.z != 0:
                raise UnexpectedValueException('Grid pos not at 0, {}'.format(keyframe.posgrille))
            if keyframe.angzgrille != 0:
                raise UnexpectedValueException('keyframe.angzgrille not 0, {}'.format(keyframe.angzgrille))

            if keyframe.fx >= 0:
                fxBytes = keyframe.fx.to_bytes(4, byteorder='little')
                fxBase = fxBytes[0]
                fxPre = fxBytes[1]
                fxPost = fxBytes[2]
                fxLight = fxBytes[3]

                fxBaseVals = ['None', 'FX_FADEIN', 'FX_FADEOUT', 'FX_BLUR']
                fxPreVals = ['None', 'FX_DREAM']
                fxPostVals = ['None', 'FX_FLASH', 'FX_APPEAR', 'FX_APPEAR2']
                fxLightVals = ['None', 'FX_LIGHT']

                if fxBase > len(fxBaseVals):
                    self.log.warning('Bad fx base {}'.format(fxBase))
                    fxBase = 0
                if fxPre > len(fxPreVals):
                    self.log.warning('Bad fx pre {}'.format(fxPre))
                    fxPre = 0
                if fxPost > len(fxPostVals):
                    self.log.warning('Bad fx post {}'.format(fxPost))
                    fxPost = 0
                if fxLight > len(fxLightVals):
                    self.log.warning('Bad fx post {}'.format(fxLight))
                    fxPost = 0

                self.log.info('FX: base {} pre {} post {} light {}'
                              .format(fxBaseVals[fxBase], fxPreVals[fxPre], fxPostVals[fxPost], fxLightVals[fxLight]))




            #if keyframe.color != 0:
            #    raise UnexpectedValueException('Unexpected color, {}'.format(keyframe.color))


            if i == 0:
                if keyframe.frame != 0:
                    raise UnexpectedValueException('First keyframe should be at frame 0')
            else:
                if keyframe.frame <= lastFrameNr:
                    raise UnexpectedValueException('Keyframes not in order')

            lastFrameNr = keyframe.frame

        self.log.debug('Initial sound: {}'.format(initialSound))

        if track.endframe != lastFrameNr:
            self.log.warning(
                'Track endframe {} does not match last keyframe frame {}'.format(track.endframe, lastFrameNr))

        return CinData(
            track=track,
            keyframes=keyframes
        )



# Copyright 2014-2021 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.

from ctypes import (
    LittleEndianStructure,
    c_char,
    c_uint16,
    c_uint32,
    c_int16,
    c_int32,
    c_float
)

from .dataCommon import SavedVec3, PolyTypeFlag

class SavedTextureVertex(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("pos",      SavedVec3),
        ("rhw",      c_float),
        ("color",    c_uint32),
        ("specular", c_uint32),
        ("tu",       c_float),
        ("tv",       c_float)
    ]

class EERIE_OLD_VERTEX(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("vert", SavedTextureVertex),
        ("v",    SavedVec3),
        ("norm", SavedVec3)
    ]

class EERIE_FACE_FTL(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("facetype", PolyTypeFlag),
        ("rgb",      c_uint32 * 3), # always zero
        ("vid",      c_uint16 * 3),
        ("texid",    c_int16),
        ("u",        c_float * 3),
        ("v",        c_float * 3),
        ("ou",       c_uint16 * 3),
        ("ov",       c_uint16 * 3),
        ("transval", c_float),
        ("norm",     SavedVec3),
        ("nrmls",    SavedVec3 * 3), # always zero
        ("temp",     c_float),
    ]

    def toTuple(self):
        return (self.facetype.asUInt,
                (self.vid[0], self.vid[1], self.vid[2]),
                self.texid,
                (self.u[0], self.u[1], self.u[2]),
                (self.v[0], self.v[1], self.v[2]),
                (self.ou[0], self.ou[1], self.ou[2]),
                (self.ov[0], self.ov[1], self.ov[2]),
                self.transval,
                (self.norm.x, self.norm.y, self.norm.z),
                self.temp)


class Texture_Container_FTL(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("name", c_char * 256)
    ]

class EERIE_GROUPLIST_FTL(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("name",     c_char * 256),
        ("origin",   c_int32),
        ("nb_index", c_int32),
        ("indexes",  c_int32),
        ("siz",      c_float)
    ]

class EERIE_ACTIONLIST_FTL(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("name",   c_char * 256),
        ("idx",    c_int32),
        ("action", c_int32),
        ("sfx",    c_int32),
    ]

class EERIE_SELECTIONS_FTL(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("name",        c_char * 64),
        ("nb_selected", c_int32),
        ("selected",    c_int32),
    ]


class ARX_FTL_PRIMARY_HEADER(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("ident",    c_char * 4),
        ("version",  c_float),
        ("checksum", c_char * 512) # In the c code this is currently not part of the header
    ]

class ARX_FTL_SECONDARY_HEADER(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("offset_3Ddata",            c_int32),
        ("offset_cylinder",          c_int32),
        ("offset_progressive_data",  c_int32),
        ("offset_clothes_data",      c_int32),
        ("offset_collision_spheres", c_int32),
        ("offset_physics_box",       c_int32)
    ]
    
class ARX_FTL_3D_DATA_HEADER(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("nb_vertex",     c_int32),
        ("nb_faces",      c_int32),
        ("nb_maps",       c_int32),
        ("nb_groups",     c_int32),
        ("nb_action",     c_int32),
        ("nb_selections", c_int32),
        ("origin",        c_int32),
        ("name",          c_char * 256)
    ]


def comp(a, b):
    for fld in a._fields_:
        attra = getattr(a, fld[0])
        attrb = getattr(b, fld[0])
        print("foo")

    return True


from collections import namedtuple
from dataclasses import dataclass

from ctypes import sizeof
import itertools

import logging
logging.basicConfig(level=logging.INFO)

from typing import Any, List


@dataclass(frozen=True)
class FtlMetadata():
    name: Any
    org: Any

@dataclass(frozen=True)
class FtlVertex():
    xyz: Any
    n: Any

@dataclass(frozen=True)
class FtlFace():
    vids: Any
    uvs: Any
    texid: Any
    facetype: Any
    transval: Any
    normal: Any

@dataclass
class FtlGroup():
    name: Any
    origin: Any
    indices: Any
    parentIndex: Any

@dataclass(frozen=True)
class FtlSelection():
    name: Any
    indices: Any

@dataclass(frozen=True)
class FtlAction():
    name: Any
    vidx: Any

@dataclass(frozen=True)
class FtlData():
    metadata: FtlMetadata
    verts: List[FtlVertex]
    faces: List[FtlFace]
    mats: List[str]
    groups: List[FtlGroup]
    actions: List[FtlAction]
    sels: List[FtlSelection]


def getFatherIndex(groups, childIndex):
    child = groups[childIndex]
    i = childIndex-1
    while i >= 0:
        group = groups[i]
        for vertexIndex in group.indices:
            if vertexIndex == child.origin:
                return i
        i-=1

    return -1

class FtlSerializer(object):
    def __init__(self):
        self.log = logging.getLogger('FtlSerializer')

    def read(self, data) -> FtlData:
        pos = 0

        primaryHeader = ARX_FTL_PRIMARY_HEADER.from_buffer_copy(data, pos)
        pos += sizeof(ARX_FTL_PRIMARY_HEADER)

        if primaryHeader.version != 0.8325700163841248:
            self.log.debug("Unexpected ftl version: %f" % primaryHeader.version)

        secondaryHeader = ARX_FTL_SECONDARY_HEADER.from_buffer_copy(data, pos)
        pos += sizeof(ARX_FTL_SECONDARY_HEADER)

        if secondaryHeader.offset_3Ddata == -1:
            self.log.error("Invalid offset to 3d data")
            return

        assert (secondaryHeader.offset_3Ddata == pos), "3D Data should follow secondary header"
        assert (secondaryHeader.offset_cylinder <= 0), "Cylinder data should be unused"
        assert (secondaryHeader.offset_progressive_data <= 0), "Progressive data should be unused"
        assert (secondaryHeader.offset_physics_box <= 0), "Physics data should be unused"

        if (secondaryHeader.offset_clothes_data > 0):
            self.log.debug("Found unused clothes data")

        if (secondaryHeader.offset_collision_spheres > 0):
            self.log.debug("Found unused collision_spheres data")

        pos = secondaryHeader.offset_3Ddata

        chunkHeader = ARX_FTL_3D_DATA_HEADER.from_buffer_copy(data, pos)
        pos += sizeof(ARX_FTL_3D_DATA_HEADER)

        metadata = FtlMetadata(name=chunkHeader.name.decode('iso-8859-1'), org=chunkHeader.origin)

        # self.log.debug("Name %s" % chunkHeader.name.decode('iso-8859-1'))  # This might be empty

        verts = []
        for i in range(chunkHeader.nb_vertex):
            vert = EERIE_OLD_VERTEX.from_buffer_copy(data, pos)
            pos += sizeof(EERIE_OLD_VERTEX)
            verts.append(FtlVertex(xyz=(vert.v.x, vert.v.y, vert.v.z), n=(vert.norm.x, vert.norm.y, vert.norm.z)))

        # debugFaces = []
        faces = []
        for i in range(chunkHeader.nb_faces):
            face = EERIE_FACE_FTL.from_buffer_copy(data, pos)
            pos += sizeof(EERIE_FACE_FTL)

            assert (face.rgb[0] == 0 and face.rgb[1] == 0 and face.rgb[2] == 0)
            assert (face.nrmls[0].x == 0 and face.nrmls[1].x == 0 and face.nrmls[2].x == 0)
            assert (face.nrmls[0].y == 0 and face.nrmls[1].y == 0 and face.nrmls[2].y == 0)
            assert (face.nrmls[0].z == 0 and face.nrmls[1].z == 0 and face.nrmls[2].z == 0)

            if face.texid < 0:
                self.log.debug("Face negative texture id %i" % i)

            oface = FtlFace((face.vid[0], face.vid[1], face.vid[2]),
                            list(zip(face.u, face.v)),
                            face.texid,
                            face.facetype.asUInt,
                            face.transval,
                            (face.norm.x, face.norm.y, face.norm.z))
            faces.append(oface)
            """
            debugFaceVids = (face.vid[0], face.vid[1], face.vid[2])
            for j, df in debugFaces:
                myVids = (df.vid[0], df.vid[1], df.vid[2])
                for p in itertools.permutations(myVids):
                    if debugFaceVids == p:
                        print("face %i %s" % (i, str(face.toTuple())))
                        print("dupl %i %s" % (j, str(df.toTuple())))
            debugFaces.append((i, face))
            """

        mats = []
        for i in range(chunkHeader.nb_maps):
            texture = Texture_Container_FTL.from_buffer_copy(data, pos)
            pos += sizeof(Texture_Container_FTL)
            mats.append(texture.name.decode('iso-8859-1'))

        temp = []
        for i in range(chunkHeader.nb_groups):
            group = EERIE_GROUPLIST_FTL.from_buffer_copy(data, pos)
            pos += sizeof(EERIE_GROUPLIST_FTL)

            # print(str(group.indexes) + ":" + str(group.nb_index) + ":" + str(group.siz))
            # assert (group.indexes == group.nb_index)
            # assert (group.siz == 0)

            temp.append((group.name.decode('iso-8859-1'), group.origin, group.nb_index))

        groups = []
        for (name, origin, count) in temp:
            vertArray = c_int32 * count
            vertsIndices = vertArray.from_buffer_copy(data, pos)
            pos += sizeof(vertsIndices)
            # 0 temporarily for parentIndex.
            groups.append(FtlGroup(name, origin, list(vertsIndices), 0))

        # parenting
        for i, group in enumerate(groups):
            father = getFatherIndex(groups, i)
            self.log.debug("group %d with father %d" % (i,father))
            group.parentIndex = father

        actions = []
        for i in range(chunkHeader.nb_action):
            action = EERIE_ACTIONLIST_FTL.from_buffer_copy(data, pos)
            pos += sizeof(EERIE_ACTIONLIST_FTL)
            actions.append(FtlAction(action.name.decode('iso-8859-1'), action.idx))

        temp = []
        for i in range(chunkHeader.nb_selections):
            selection = EERIE_SELECTIONS_FTL.from_buffer_copy(data, pos)
            pos += sizeof(EERIE_SELECTIONS_FTL)
            temp.append((selection.name.decode('iso-8859-1'), selection.nb_selected))

        sels = []
        for (name, count) in temp:
            vertArray = c_int32 * count
            vertsIndices = vertArray.from_buffer_copy(data, pos)
            pos += sizeof(vertsIndices)
            sels.append(FtlSelection(name, list(vertsIndices)))

        return FtlData(
            metadata=metadata,
            verts=verts,
            faces=faces,
            mats=mats,
            groups=groups,
            actions=actions,
            sels=sels
        )

    def write(self, data: FtlData) -> bytearray:

        result = bytearray()

        primaryHeader = ARX_FTL_PRIMARY_HEADER()
        primaryHeader.ident = b'FTL'
        primaryHeader.version = 0.8325700163841248
        result += bytearray(primaryHeader)

        secondaryHeader = ARX_FTL_SECONDARY_HEADER()
        secondaryHeader.offset_3Ddata = sizeof(ARX_FTL_PRIMARY_HEADER) + sizeof(ARX_FTL_SECONDARY_HEADER)

        secondaryHeader.offset_cylinder = -1
        secondaryHeader.offset_progressive_data = -1
        secondaryHeader.offset_clothes_data = -1
        secondaryHeader.offset_collision_spheres = -1
        secondaryHeader.offset_physics_box = -1
        result += bytearray(secondaryHeader)

        chunkHeader = ARX_FTL_3D_DATA_HEADER()
        chunkHeader.nb_vertex = len(data.verts)
        chunkHeader.nb_faces = len(data.faces)
        chunkHeader.nb_maps = len(data.mats)
        chunkHeader.nb_groups = len(data.groups)
        chunkHeader.nb_action = len(data.actions)
        chunkHeader.nb_selections = len(data.sels)

        chunkHeader.origin = data.metadata.org
        chunkHeader.name = data.metadata.name.encode('iso-8859-1')
        result += bytearray(chunkHeader)

        self.log.debug("Verts start: %s" % hex(len(result)))

        for v in data.verts:
            vertex = EERIE_OLD_VERTEX()
            vertex.v = SavedVec3(v.xyz[0], v.xyz[1], v.xyz[2])
            vertex.norm = SavedVec3(v.n[0], v.n[1], v.n[2])
            result += bytearray(vertex)

        self.log.debug("Faces start: %s" % hex(len(result)))

        for f in data.faces:
            face = EERIE_FACE_FTL()
            face.vid = (c_uint16 * 3)(*f.vids)
            face.u, face.v = zip(*f.uvs)
            
            if f[2] < len(data.mats):
                face.texid = f[2]
            else:
                face.texid = -1
            
            #face.texid = f[2]
            face.facetype.asUInt = f[3]
            face.transval = f[4]
            # face.norm.x = f.normal[0]
            # face.norm.y = f.normal[1]
            # face.norm.z = f.normal[2]
            result += bytearray(face)

        self.log.debug("Mats start: %s" % hex(len(result)))

        for m in data.mats:
            texture = Texture_Container_FTL()
            texture.name = m.encode('iso-8859-1')
            result += bytearray(texture)

        self.log.debug("Groups start: %s" % hex(len(result)))

        for g in data.groups:
            group = EERIE_GROUPLIST_FTL()
            group.name = g[0].encode('iso-8859-1')
            group.origin = g[1]
            group.nb_index = len(g[2])
            result += bytearray(group)

        for g in data.groups:
            VertArray = c_int32 * len(g[2])
            array = (VertArray)(*g[2])
            result += bytearray(array)

        self.log.debug("Actions start: %s" % hex(len(result)))

        for a in data.actions:
            action = EERIE_ACTIONLIST_FTL()
            action.name = a[0].encode('iso-8859-1')
            action.idx = a[1]
            result += bytearray(action)

        self.log.debug("Sels start: %s" % hex(len(result)))

        for s in data.sels:
            selection = EERIE_SELECTIONS_FTL()
            selection.name = s[0].encode('iso-8859-1')
            selection.nb_selected = len(s[1])
            result += bytearray(selection)

        for s in data.sels:
            VertArray = c_int32 * len(s[1])
            array = (VertArray)(*s[1])
            result += bytearray(array)

        return result



# Copyright 2014-2019 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.

import os
import platform
import ctypes
import logging

log = logging.getLogger('arx addon')

import sys
def getCurrentMachine():
    if sys.maxsize > 2**32:
        return 'x64'
    else:
        return 'x86'

import struct
def getDllMachine(dllFile):
    with open(dllFile, "rb") as f:
        f.seek(60)
        e_lfanew, = struct.unpack("<l", f.read(4))
        f.seek(e_lfanew)
        pemagic = f.read(4)
        if pemagic == b'PE\x00\x00':
            machine, = struct.unpack("<H", f.read(2))
            if machine == 0x8664:
                return 'x64'
            elif machine == 0x014c:
                return 'x86'
            else:
                raise RuntimeError("Unknown machine type in PE header" + str(machine))
        else:
            raise RuntimeError("PE magic not found" + str(pemagic))

def checkDll(dllFile):
    a = getDllMachine(dllFile)
    b = getCurrentMachine()
    if a != b:
        raise RuntimeError("Can not load '"+a+"' '" + dllFile + "' dll in '"+b+"' Blender, please use matching dll")

class ArxIO(object):
    messageBufferSize = 512

    def __init__(self):
        if platform.system() == "Windows":
            libPaths = [ os.path.realpath(__file__ + "\..\ArxIO.dll") ]
            checkDll(libPaths[0])
        else:
            libPaths = [
                os.path.realpath(__file__ + "/../libArxIO.so.0"),
                os.path.realpath(__file__ + "/../libArxIO.so"),
                "libArxIO.so.0"
            ]
        
        self.lib = None
        lastException = None
        for libPath in libPaths:
            if os.path.isfile(libPath):
                try:
                    self.lib = ctypes.cdll.LoadLibrary(libPath)
                    break
                except Exception as e:
                    lastException = e
                    continue
        if self.lib is None:
            raise Exception('could not load the ArxIO library from: [\n' + ',\n'.join(libPaths) + '\n]\nLoadLibrary Exception: ' + str(lastException))
        self.lib.ArxIO_init()

    def getError(self):
        errorBuffer = ctypes.create_string_buffer(self.messageBufferSize)
        self.lib.ArxIO_getError(errorBuffer, self.messageBufferSize)
        return RuntimeError(errorBuffer.value)
        
    def printLog(self):
        readMore = True
        while readMore:
            logBuffer = ctypes.create_string_buffer(self.messageBufferSize)
            result = self.lib.ArxIO_getLogLine(logBuffer, self.messageBufferSize)
            log.info("Lib " + logBuffer.value.decode("utf-8"))
            if result == 0:
                readMore = False;

    def unpack(self, data):
        cData = ctypes.create_string_buffer(data, len(data))
        
        out = ctypes.POINTER(ctypes.c_char)()
        outSize = ctypes.c_size_t()
        # FIXME error handling
        self.lib.ArxIO_unpack_alloc(cData, len(cData), ctypes.byref(out), ctypes.byref(outSize))
        
        result = ctypes.create_string_buffer(outSize.value)
        ctypes.memmove(result, out, outSize.value)
        
        self.lib.ArxIO_unpack_free(out)
        
        return result



# Copyright 2014-2021 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.

from ctypes import (
    LittleEndianStructure,
    c_char,
    c_int32,
    c_float,
    c_ubyte
)

from .dataCommon import (
    SavedVec3,
    SavedColor
)

class DANAE_LLF_HEADER(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("version",          c_float),
        ("ident",            c_char * 16),
        ("lastuser",         c_char * 256),
        ("time",             c_int32),
        ("nb_lights",        c_int32),
        ("nb_Shadow_Polys",  c_int32),
        ("nb_IGNORED_Polys", c_int32),
        ("nb_bkgpolys",      c_int32),
        ("pad",              c_int32 * 256),
        ("fpad",             c_float * 256),
        ("cpad",             c_char * 4096),
        ("bpad",             c_int32 * 256)
    ]

class DANAE_LS_LIGHT(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("pos",          SavedVec3),
        ("rgb",          SavedColor),
        ("fallstart",    c_float),
        ("fallend",      c_float),
        ("intensity",    c_float),
        ("i",            c_float),
        ("ex_flicker",   SavedColor),
        ("ex_radius",    c_float),
        ("ex_frequency", c_float),
        ("ex_size",      c_float),
        ("ex_speed",     c_float),
        ("ex_flaresize", c_float),
        ("fpadd",        c_float * 24),
        ("extras",       c_int32),
        ("lpadd",        c_int32 * 31)
    ]

class DANAE_LS_LIGHTINGHEADER(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("nb_values", c_int32),
        ("ViewMode",  c_int32),
        ("ModeLight", c_int32),
        ("pad",       c_int32)
    ]

class SavedColorBGRA(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("b", c_ubyte),
        ("g", c_ubyte),
        ("r", c_ubyte),
        ("a", c_ubyte)
    ]

import logging
from ctypes import sizeof

from collections import namedtuple

LlfData = namedtuple('LlfData', ['lights', 'levelLighting'])

class LlfSerializer(object):
    def __init__(self, ioLib):
        self.log = logging.getLogger('LlfSerializer')
        self.ioLib = ioLib
    
    def read(self, fileName) -> LlfData:
        f = open(fileName, "rb")
        compressedData = f.read()
        f.close()
        self.log.info("Loaded %i bytes from file %s" % (len(compressedData), fileName))
        
        data = self.ioLib.unpack(compressedData)
        
        pos = 0
        
        llfHeader = DANAE_LLF_HEADER.from_buffer_copy(data, pos)
        pos += sizeof(DANAE_LLF_HEADER)
        
        LightsList = DANAE_LS_LIGHT * llfHeader.nb_lights
        lights = LightsList.from_buffer_copy(data, pos)
        pos += sizeof(LightsList)

        lightingHeader = DANAE_LS_LIGHTINGHEADER.from_buffer_copy(data, pos)
        pos += sizeof(DANAE_LS_LIGHTINGHEADER)

        VertexColors = SavedColorBGRA * lightingHeader.nb_values
        levelLighting = VertexColors.from_buffer_copy(data, pos)
        pos += sizeof(VertexColors)

        if len(data) - pos != 0:
            self.log.info("Unexpected data at end of file")

        return LlfData(
            lights=lights,
            levelLighting=levelLighting
        )



# Copyright 2015-2020 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.

import importlib
import sys
import os
import logging
import itertools

if "bpy" in locals():
    importlib.reload(sys.modules["arx_addon.arx_io_animation"])
    importlib.reload(sys.modules["arx_addon.arx_io_area"])
    importlib.reload(sys.modules["arx_addon.arx_io_material"])
    importlib.reload(sys.modules["arx_addon.arx_io_model"])
    importlib.reload(sys.modules["arx_addon.arx_io_test_roundtrip"])
    importlib.reload(sys.modules["arx_addon.arx_io_util"])
    importlib.reload(sys.modules["arx_addon.dataCommon"])
    importlib.reload(sys.modules["arx_addon.dataDlf"])
    importlib.reload(sys.modules["arx_addon.dataFtl"])
    importlib.reload(sys.modules["arx_addon.dataFts"])
    importlib.reload(sys.modules["arx_addon.dataLlf"])
    importlib.reload(sys.modules["arx_addon.dataTea"])
    importlib.reload(sys.modules["arx_addon.files"])
    importlib.reload(sys.modules["arx_addon.lib"])

import bpy
import bmesh

from .lib import ArxIO
from .files import *

from .arx_io_animation import ArxAnimationManager
from .arx_io_area import ArxSceneManager
from .arx_io_model import ArxObjectManager
from .arx_io_test_roundtrip import test_export_of_current_scene_objects

log = logging.getLogger('ArxAddon')

def triangulate(bm):
    while True:
        nonTris = [f for f in bm.faces if len(f.verts) > 3]
        if nonTris:
            nt = nonTris[0]
            pivotLoop = nt.loops[0]
            allVerts = nt.verts
            vert1 = pivotLoop.vert
            wrongVerts = [vert1, pivotLoop.link_loop_next.vert, pivotLoop.link_loop_prev.vert]
            bmesh.utils.face_split(nt, vert1, [v for v in allVerts if v not in wrongVerts][0])

            for seq in [bm.verts, bm.faces, bm.edges]:
                seq.index_update()
        else:
            break
    return bm


def strip_wires(bm):
    [bm.verts.remove(v) for v in bm.verts if v.is_wire]
    [bm.verts.remove(v) for v in bm.verts if not v.link_faces[:]]
    [bm.edges.remove(e) for e in bm.edges if not e.link_faces[:]]
    [bm.faces.remove(f) for f in bm.faces if len(f.edges) < 3]
    for seq in [bm.verts, bm.faces, bm.edges]: seq.index_update()
    return bm



class ArxAssetManager(object):
    def __init__(self, arxFiles, objectManager, sceneManager):
        self.log = logging.getLogger('ArxAssetManager')
        self.log.setLevel(logging.DEBUG)
        
        self.arxFiles = arxFiles
        self.objectManager = objectManager
        self.sceneManager = sceneManager

    def importAllModels(self, context):
        sortedModels = sorted(self.arxFiles.models.data.items())
        
        for i, (key, val) in enumerate(sortedModels):
            """
            sceneName = "m:" + '/'.join(key)
            
            scene = bpy.data.scenes.get(sceneName)
            
            if scene is None:
                self.log.info("Creating new scene [{}]".format(sceneName))
                scene = bpy.data.scenes.new(name=sceneName)
                scene.unit_settings.system = 'METRIC'
                scene.unit_settings.scale_length = 0.01
            else:
                self.log.info("Scene [{}] already exists".format(sceneName))
                continue
            """
            scene = None

            import_file = os.path.join(val.path, val.model)
            self.objectManager.loadFile(context, import_file, scene, False)


class ArxAddon(object):
    def __init__(self, dataPath, allowLibFallback):
        self.log = logging.getLogger('Arx Addon')
        
        try:
            ioLib = ArxIO()
        except Exception as e:
            if not allowLibFallback:
                raise e
            else:
                self.log.error("Failed to load native io library, using slow fallback. Exception: " + str(e));
                
                from .naivePkware import decompress_ftl
                class ArxIOFallback(object):
                    def __init__(self):
                        pass
                    def unpack(self, data):
                        return decompress_ftl(data)
                        
                ioLib = ArxIOFallback()
        
        self.arxFiles = ArxFiles(dataPath)
        self.arxFiles.updateAll()
        
        self.objectManager = ArxObjectManager(ioLib, dataPath)
        self.sceneManager = ArxSceneManager(ioLib, dataPath, self.arxFiles, self.objectManager)
        self.animationManager = ArxAnimationManager()
        self.assetManager = ArxAssetManager(self.arxFiles, self.objectManager, self.sceneManager)
    
    def testModelExport(self):
        
        test_export_of_current_scene_objects(self.assetManager)

# ======================================================================================================================

g_arxAddon = None

def arxAddonReload():
    log.info("invalidating")
    global g_arxAddon
    g_arxAddon = None

def getAddon(context):
    global g_arxAddon
    if not g_arxAddon:
        log.info("creating")
        addon_prefs = context.preferences.addons[__package__].preferences
        g_arxAddon = ArxAddon(addon_prefs.arxAssetPath, addon_prefs.arxAllowLibFallback)
    
    return g_arxAddon



# Copyright 2019-2020 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.

import logging

import bpy
import bmesh
from math import radians
from mathutils import Vector, Matrix

from .dataDlf import DlfSerializer, DlfData
from .dataFts import FtsSerializer
from .dataLlf import LlfSerializer

from .arx_io_material import createMaterial
from .arx_io_util import arx_pos_to_blender_for_model, ArxException

correctionMatrix = \
    Matrix.Rotation(radians(180), 4, 'Z') @ \
    Matrix.Rotation(radians(-90), 4, 'X')

class ArxSceneManager(object):
    def __init__(self, ioLib, dataPath, arxFiles, objectManager):
        self.log = logging.getLogger('ArxSceneManager')
        self.dlfSerializer = DlfSerializer(ioLib)
        self.ftsSerializer = FtsSerializer(ioLib)
        self.llfSerializer = LlfSerializer(ioLib)
        self.dataPath = dataPath
        self.arxFiles = arxFiles
        self.objectManager = objectManager

    def importScene(self, context, scene, area_id):
        self.log.info('Importing Area: {}'.format(area_id))
        
        area_files = self.arxFiles.levels.levels[area_id]
        
        if area_files.dlf is None:
            self.log.error("dlf file not found")
            return
        if area_files.fts is None:
            self.log.error("fts file not found")
            return
        if area_files.llf is None:
            self.log.error("llf file not found")
            return
        
        dlfData = self.dlfSerializer.readContainer(area_files.dlf)
        ftsData = self.ftsSerializer.read_fts_container(area_files.fts)
        llfData = self.llfSerializer.read(area_files.llf)

        # bpy.types.Material.Shader_Name = bpy.props.StringProperty(name='Group Name')

        # Create materials
        mappedMaterials = []
        idx = 0
        for tex in ftsData.textures:
            mappedMaterials.append((idx, tex.tc, createMaterial(self.dataPath, tex.fic.decode('iso-8859-1'))))
            idx += 1

        # Create mesh
        bm = self.AddSceneBackground(ftsData.cells, llfData.levelLighting, mappedMaterials)
        mesh = bpy.data.meshes.new(scene.name + "-mesh")
        bm.to_mesh(mesh)
        bm.free()

        # Create background object
        obj = bpy.data.objects.new(scene.name + "-background", mesh)
        scene.collection.objects.link(obj)
        # scn.objects.active = obj
        # obj.select = True

        # Create materials
        for idx, tcId, mat in mappedMaterials:
            obj.data.materials.append(mat)

        self.AddScenePathfinderAnchors(scene, ftsData.anchors)
        self.AddScenePortals(scene, ftsData)
        self.AddSceneLights(scene, llfData, ftsData.sceneOffset)
        self.AddSceneObjects(scene, dlfData, ftsData.sceneOffset)
        self.add_scene_map_camera(scene)

    def AddSceneBackground(self, cells, levelLighting, mappedMaterials):
        bm = bmesh.new()
        uvLayer = bm.loops.layers.uv.verify()
        colorLayer = bm.loops.layers.color.new("light-color")

        


        vertexIndex = 0
        for z in cells:
            for cell in z:

                if cell is None:
                    continue

                for face in cell:

                    if face.type.POLY_QUAD:
                        to = 4
                    else:
                        to = 3

                    tempVerts = []
                    for i in range(to):
                        pos = [face.v[i].ssx, face.v[i].sy, face.v[i].ssz]
                        uv = [face.v[i].stu, 1 - face.v[i].stv]
                        intCol = levelLighting[vertexIndex]
                        floatCol = (intCol.r / 255.0, intCol.g / 255.0, intCol.b / 255.0, intCol.a / 255.0)
                        tempVerts.append((pos, uv, floatCol))
                        vertexIndex += 1

                    # Switch the vertex order
                    if face.type.POLY_QUAD:
                        tempVerts[2], tempVerts[3] = tempVerts[3], tempVerts[2]

                    vertIdx = []
                    for i in tempVerts:
                        vertIdx.append(bm.verts.new(arx_pos_to_blender_for_model(i[0])))

                    bmFace = bm.faces.new(vertIdx)

                    if face.tex != 0:
                        matIdx = next((x for x in mappedMaterials if x[1] == face.tex), None)

                        if matIdx is not None:
                            bmFace.material_index = matIdx[0]
                        else:
                            self.log.info("Matrial id not found %i" % face.tex)

                    for i, loop in enumerate(bmFace.loops):
                        loop[uvLayer].uv = tempVerts[i][1]
                        loop[colorLayer] = tempVerts[i][2]

        bm.verts.index_update()
        bm.edges.index_update()
        #bm.transform(correctionMatrix)
        return bm
    
    def AddScenePathfinderAnchors(self, scene, anchors):
        
        bm = bmesh.new()
        
        bVerts = []
        for anchor in anchors:
            bVerts.append(bm.verts.new(arx_pos_to_blender_for_model(anchor[0])))
        
        bm.verts.index_update()
        
        for i, anchor in enumerate(anchors):
            for edge in anchor[1]:
                #TODO this is a hack
                try:
                    bm.edges.new((bVerts[i], bVerts[edge]));
                except ValueError:
                    pass
        
        #bm.transform(correctionMatrix)
        mesh = bpy.data.meshes.new(scene.name + '-anchors-mesh')
        bm.to_mesh(mesh)
        bm.free()
        obj = bpy.data.objects.new(scene.name + '-anchors', mesh)
        # obj.draw_type = 'WIRE'
        # obj.show_x_ray = True
        scene.collection.objects.link(obj)

    def AddScenePortals(self, scene, data):
        portals_col = bpy.data.collections.new(scene.name + '-portals')
        scene.collection.children.link(portals_col)

        for portal in data.portals:
            bm = bmesh.new()

            tempVerts = []
            for vertex in portal.poly.v:
                pos = [vertex.pos.x, vertex.pos.y, vertex.pos.z]
                tempVerts.append(pos)

            # Switch the vertex order
            tempVerts[2], tempVerts[3] = tempVerts[3], tempVerts[2]

            bVerts = []
            for i in tempVerts:
                bVerts.append(bm.verts.new(arx_pos_to_blender_for_model(i)))

            bm.faces.new(bVerts)
            #bm.transform(correctionMatrix)
            mesh = bpy.data.meshes.new(scene.name + '-portal-mesh')
            bm.to_mesh(mesh)
            bm.free()
            obj = bpy.data.objects.new(scene.name + '-portal', mesh)
            obj.display_type = 'WIRE'
            obj.display.show_shadows = False
            # obj.show_x_ray = True
            # obj.hide = True
            #obj.parent_type = 'OBJECT'
            #obj.parent = groupObject
            portals_col.objects.link(obj)

    def AddSceneLights(self, scene, llfData, sceneOffset):
        lights_col = bpy.data.collections.new(scene.name + '-lights')
        scene.collection.children.link(lights_col)

        for index, light in enumerate(llfData.lights):
            light_name = scene.name + '-light_' + str(index).zfill(4)

            lampData = bpy.data.lights.new(name=light_name, type='POINT')
            lampData.color = (light.rgb.r, light.rgb.g, light.rgb.b)
            lampData.use_custom_distance = True
            lampData.cutoff_distance = light.fallend
            lampData.energy = light.intensity * 1000 # TODO this is a guessed factor

            obj = bpy.data.objects.new(name=light_name, object_data=lampData)
            lights_col.objects.link(obj)
            abs_loc = Vector(sceneOffset) + Vector([light.pos.x, light.pos.y, light.pos.z])
            obj.location = arx_pos_to_blender_for_model(abs_loc)


    def AddSceneObjects(self, scene, dlfData: DlfData, sceneOffset):
        entities_col = bpy.data.collections.new(scene.name + '-entities')
        scene.collection.children.link(entities_col)

        for e in dlfData.entities:
            
            legacyPath = e.name.decode('iso-8859-1').replace("\\", "/").lower().split('/')
            objectId = '/'.join(legacyPath[legacyPath.index('interactive') + 1 : -1])

            entityId = objectId + "_" + str(e.ident).zfill(4)
            self.log.info("Creating entity [{}]".format(entityId))

            proxyObject = bpy.data.objects.new(name='e:' + entityId, object_data=None)
            entities_col.objects.link(proxyObject)

            object_col = bpy.data.collections.get(objectId)
            if object_col:
                proxyObject.instance_type = 'COLLECTION'
                proxyObject.instance_collection = object_col
            else:
                proxyObject.show_name = True
                proxyObject.empty_display_type = 'ARROWS'
                proxyObject.empty_display_size = 20 #cm
                #self.log.info("Object not found: [{}]".format(objectId))

            pos = Vector(sceneOffset) + Vector([e.pos.x, e.pos.y, e.pos.z])
            proxyObject.location = arx_pos_to_blender_for_model(pos)

            # FIXME proper rotation conversion
            #proxyObject.rotation_mode = 'YXZ'
            proxyObject.rotation_euler = [radians(e.angle.a), radians(e.angle.g), radians(e.angle.b)]

    def add_scene_map_camera(self, scene):
        """Grid size is 160x160m"""
        cam = bpy.data.cameras.new('Map Camera')
        cam.type = 'ORTHO'
        cam.ortho_scale = 16000
        cam.clip_start = 100 # 1m
        cam.clip_end = 20000 # 200m
        cam.show_name = True
        cam_obj = bpy.data.objects.new('Map Camera', cam)
        cam_obj.location = Vector((8000.0, 8000.0, 5000.0))
        scene.collection.objects.link(cam_obj)

        scene.render.engine = 'BLENDER_EEVEE'
        scene.render.resolution_x = 1000
        scene.render.resolution_y = 1000


# Copyright 2014-2020 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.

bl_info = {
    "name": "Arx Libertatis Addon",
    "author": "Arx Libertatis Team",
    "version": (0, 0, 2),
    "blender": (2, 80, 0),
    "location": "",
    "description": "Addon for managing Arx Libertatis assets",
    "warning": "This is alpha software, read the documentation, keep your original blend files!",
    "wiki_url": "https://wiki.arx-libertatis.org/Blender",
    "tracker_url": "https://bugs.arx-libertatis.org/arx/issues/new",
    "category": "Import-Export",
    "support": "COMMUNITY",
}

import imp

try:
    imp.find_module('bpy')
    blenderFound = True
except ImportError:
    blenderFound = False

print("Running inside blender: " + str(blenderFound))

if blenderFound:
    import importlib
    if "main" in locals():
        importlib.reload(main)

    from .main import register

    def register():
        main.register();

    def unregister():
        main.unregister();





# Copyright 2019-2021 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.

import os
import itertools
import logging
import math

import bpy
import bmesh
from mathutils import Vector

from .arx_io_material import createMaterial
from .arx_io_util import (
    arx_pos_to_blender_for_model,
    blender_pos_to_arx,
    ArxException
)
from .files import splitPath
from .dataFtl import (
    FtlMetadata,
    FtlVertex,
    FtlFace,
    FtlGroup,
    FtlSelection,
    FtlAction,
    FtlData,
    FtlSerializer
)

class ArxObjectManager(object):
    def __init__(self, ioLib, dataPath):
        self.log = logging.getLogger('ArxObjectManager')
        self.ioLib = ioLib
        self.dataPath = dataPath
        self.ftlSerializer = FtlSerializer()

    def validateAssetDirectory(self):
        if not os.path.isdir(self.dataPath):
            raise ArxException("Arx assert directory path [" + self.dataPath + "] is not a directory")
        
        if not os.path.exists(self.dataPath):
            raise ArxException("Arx assert directory path [" + self.dataPath + "] does not exist")

    def validateObjectName(self, name):
        if len(name.encode('utf-8')) > 63:
            raise ArxException("Name ["+name+"] too long to be used as blender object name")

    def analyzeFaceData(self, faceData):
        # find weird face data
        facesToDrop = []

        seenFaceVerts = {}
        for i, face in enumerate(faceData):
            for foo in itertools.permutations(face.vids):
                if foo in seenFaceVerts:
                    seen = seenFaceVerts[foo]
                    facesToDrop.append(seen)
                    break

            seenFaceVerts[face.vids] = i

        if len(facesToDrop) > 0:
            self.log.debug("Dropping faces: " + str(facesToDrop))

        return facesToDrop

    def createBmesh(self, context, vertData, faceData) -> bmesh.types.BMesh:

        facesToDrop = self.analyzeFaceData(faceData)

        bm = bmesh.new()

        arxFaceType = bm.faces.layers.int.new('arx_facetype')
        arxTransVal = bm.faces.layers.float.new('arx_transval')

        uvData = bm.loops.layers.uv.verify()
        for i, v in enumerate(vertData):
            vertex = bm.verts.new(arx_pos_to_blender_for_model(v.xyz))
            vertex.normal = v.n
            vertex.index = i

        bm.verts.ensure_lookup_table()

        for i, f in enumerate(faceData):

            if i in facesToDrop:
                continue

            # if f.texid < 0:
            #    continue

            faceVerts = [bm.verts[v] for v in f.vids]
            evDict = {}
            try:
                face = bm.faces.new(faceVerts)
                face.index = i
            except ValueError:
                self.log.debug("Extra face")

            face.normal = Vector(f.normal)

            if f.texid >= 0:
                face.material_index = f.texid

            face[arxFaceType] = f.facetype
            face[arxTransVal] = f.transval

            for j, loop in enumerate(face.loops):
                u, v = f.uvs[j]
                loop[uvData].uv = u, 1.0 - v

        bm.faces.ensure_lookup_table()

        bm.edges.index_update()
        bm.edges.ensure_lookup_table()

        return bm

    def createObject(self, context, bm, data:FtlData, canonicalId, collection) -> bpy.types.Object:

        idString = "/".join(canonicalId);
        self.validateObjectName(idString)
        mesh = bpy.data.meshes.new(idString)
        bm.to_mesh(mesh)

        #armatureObj = self.createArmature(canonicalId, bm, data.groups)

        bm.free()

        if bpy.context.active_object:
            bpy.ops.object.mode_set(mode='OBJECT')

        obj = bpy.data.objects.new(name=mesh.name, object_data=mesh)
        #armatureObj.parent = obj # this created a dependency recursion, so i commented it

        obj['arx.ftl.name'] = data.metadata.name
        obj['arx.ftl.org'] = data.metadata.org

        for i, g in enumerate(data.groups):
            grp = obj.vertex_groups.new(name="grp:" + str(i).zfill(2) + ":" + g.name)
            # XXX add the origin to the group ?
            for v in g.indices:
                grp.add([v], 0.0, 'ADD')

        for i, s in enumerate(data.sels):
            grp = obj.vertex_groups.new(name="sel:" + str(i).zfill(2) + ":" + s.name)

            for v in s.indices:
                grp.add([v], 0.0, 'ADD')

        for a in data.actions:
            action = bpy.data.objects.new(a.name, None)
            action.parent = obj
            action.parent_type = 'VERTEX'
            action.parent_vertices = [a.vidx, 0, 0]  # last two are ignored
            action.show_name = True
            collection.objects.link(action)

            if a.name.lower().startswith("hit_"):
                radius = int(a.name[4:])
                action.empty_display_type = 'SPHERE'
                action.empty_display_size = radius
                action.lock_rotation = [True, True, True]
                action.lock_scale = [True, True, True]
            else:
                action.scale = [3, 3, 3]
                
            
        #armatureModifier = obj.modifiers.new(type='ARMATURE', name="Skeleton")
        #armatureModifier.object = armatureObj


        # for toDeSel in scene.objects: #deselect all first just to be sure
        #    toDeSel.select = False

        collection.objects.link(obj)
        #FIXME properly bind bones to mesh via vertex groups
        #obj.select = True
        #armatureObj.select = True
        #bpy.context.scene.objects.active = armatureObj #select both the mesh and armature and set armature active
        #bpy.ops.object.parent_set(type='ARMATURE_AUTO') # ctrl+p and automatic weights
        #bpy.context.scene.objects.active = obj
        #armatureObj.select = False

        #bpy.ops.object.mode_set(mode='EDIT', toggle=False)  # initialises UVmap correctly


        # TODO 2.8
        # mesh.uv_textures.new()
        for m in data.mats:
            mat = createMaterial(self.dataPath, m)
            obj.data.materials.append(mat)
            
        #bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
        return obj

    def createArmature(self, canonicalId, bm, groups) -> bpy.types.Object:

        amtname = "/".join(canonicalId)
        origin = (0, 0, 0)

        # Create armature and object
        amt = bpy.data.armatures.new(amtname)
        amt.draw_type = 'WIRE'
        amtobject = bpy.data.objects.new(amtname + "-amt", amt)
        # amtobject.show_x_ray = True
        amtobject.location = origin
        # ob.show_name = True

        # Link object to scene and make active
        bpy.context.scene.collection.link(amtobject)
        bpy.context.scene.collection.active = amtobject
        amtobject.select = True

        bpy.ops.object.mode_set(mode='EDIT')

        for i, (name, origin, indices, parentIndex) in enumerate(groups):
            bGrpName = "grp:" + str(i).zfill(2) + ":" + name

            bone = amt.edit_bones.new(bGrpName)
            bone.head = bm.verts[origin].co
            bone.tail = bm.verts[origin].co + Vector((0, -5, 0))
            bone["OriginVertex"] = origin

        editBonesArray = amt.edit_bones.values()
        for i, (name, origin, indices, parentIndex) in enumerate(groups):
            if parentIndex>=0:
                bone = editBonesArray[i]
                bone.parent = editBonesArray[parentIndex]

        bpy.ops.object.mode_set(mode='OBJECT')

        return amtobject

    def loadFile_data(self, filePath) -> FtlData:
        self.validateAssetDirectory();
        
        self.log.debug("Loading file: %s" % filePath)

        with open(filePath, "rb") as f:
            data = f.read()
            if data[:3] == b"FTL":
                unpacked = data
                self.log.debug("Loaded %i unpacked bytes from file %s" % (len(data), filePath))
            else:
                unpacked = self.ioLib.unpack(data)
                with open(filePath + ".unpack", "wb") as f:
                    f.write(unpacked)
                    self.log.debug("Written unpacked ftl")
                self.log.debug("Loaded %i packed bytes from file %s" % (len(data), filePath))

        ftlData = self.ftlSerializer.read(unpacked)
        return ftlData


    def loadFile(self, context, filePath, scene, import_tweaks: bool) -> bpy.types.Object:
        ftlData = self.loadFile_data(filePath)

        # create canonical id
        objPath = os.path.join(self.dataPath, "game/graph/obj3d/interactive")
        relPath = os.path.relpath(filePath, objPath)
        split = splitPath(relPath)
        canonicalId = split[:-1]

        # build the collection tree
        parent_collection = context.scene.collection
        idLen = len(canonicalId)
        for i in range(1, idLen + 1):
            prefix_str = "/".join(canonicalId[:i])
            if prefix_str in bpy.data.collections:
                parent_collection = bpy.data.collections[prefix_str]
            else:
                col = bpy.data.collections.new(prefix_str)
                parent_collection.children.link(col)
                parent_collection = col


        object_id_string = "/".join(canonicalId)

        self.log.debug("Canonical ID: %s" % object_id_string)

        #collection = bpy.data.collections.new(object_id_string)
        #context.scene.collection.children.link(collection)

        bm = self.createBmesh(context, ftlData.verts, ftlData.faces)
        obj = self.createObject(context, bm, ftlData, canonicalId, parent_collection)

        if import_tweaks:
            file_dir = os.path.dirname(filePath)
            tweak_dir = os.path.join(file_dir, 'tweaks')
            if os.path.isdir(tweak_dir):
                for rel_tweak_file in os.listdir(tweak_dir):
                    abs_tweak_file = os.path.join(tweak_dir, rel_tweak_file)
                    tweak_data = self.loadFile_data(abs_tweak_file)
                    tweak_mesh = self.createBmesh(context, tweak_data.verts, tweak_data.faces)

                    tweak_obj_id = canonicalId.copy()
                    tweak_obj_id.append(os.path.splitext(rel_tweak_file)[0])

                    tweak_obj = self.createObject(context, tweak_mesh, tweak_data, tweak_obj_id, parent_collection)
                    tweak_obj.parent = obj


        return obj

    # =============================================================

    def toFtlData(self, obj) -> FtlData:

        #objs = [o for o in bpy.data.objects if o.type == 'MESH' and not o.hide]

        #if not objs:
        #    self.log.info("Nothing to export")

        #obj = objs[0]

        canonicalId = obj.name.split("/")
        self.log.debug("Exporting Canonical Id: %s" % str(canonicalId))

        # obj.update_from_editmode()

        bm = bmesh.new()
        bm.from_mesh(obj.data)
        bm.normal_update()

        arxFaceType = bm.faces.layers.int.get('arx_facetype')
        if not arxFaceType:
            raise ArxException("Mesh is missing arx specific data layer: " + 'arx_facetype')
        
        arxTransVal = bm.faces.layers.float.get('arx_transval')
        if not arxTransVal:
            raise ArxException("Mesh is missing arx specific data layer: " + 'arx_transval')
        
        if not (math.isclose(obj.location[0], 0.0) and math.isclose(obj.location[1], 0.0) and math.isclose(obj.location[2], 0.0)):
            raise ArxException("Object is moved, please apply the location to the vertex positions")
        
        if not (math.isclose(obj.rotation_euler[0], 0.0) and math.isclose(obj.rotation_euler[1], 0.0) and math.isclose(obj.rotation_euler[2], 0.0)):
            raise ArxException("Object is rotated, please apply the rotation to the vertex positions")
        
        if not (math.isclose(obj.scale[0], 1.0) and math.isclose(obj.scale[1], 1.0) and math.isclose(obj.scale[2], 1.0)):
            raise ArxException("Object is scaled, please apply the scale to the vertex positions")
        

        for f in bm.faces:
            if len(f.loops) > 3:
                raise ArxException("Face with more than 3 vertices found")

        for o in bpy.data.objects:
            for ms in o.material_slots:
                if not ms.material.name.endswith('-mat'):
                    raise ArxException("Material slot names must end with: " + '-mat')
                    
        
        # Validate child empties representing action points
        for child in bpy.data.objects:
            if child.type == 'EMPTY' and child.parent == obj:
                nameParts = child.name.split(".")
                name = nameParts[0].lower()
                print(nameParts)
                if name.startswith("hit_"):
                    pass
                elif name == "view_attach":
                    pass
                elif name == "primary_attach":
                    pass
                elif name == "left_attach":
                    pass
                elif name == "weapon_attach":
                    pass
                elif name == "secondary_attach":
                    pass
                elif name == "fire":
                    pass
                else:
                    self.log.warn("Unexpected child empty: {}".format(nameParts))
                    
        
        # TODO validate the mesh

        verts = []
        
        # First add all mesh vertices
        for v in bm.verts:
            vertexNormals = [sc.normal for sc in bm.verts if sc.co == v.co]
            normal = Vector((0.0, 0.0, 0.0))
            for n in vertexNormals:
                normal += v.normal
            normal.normalize()
            verts.append(FtlVertex(
                blender_pos_to_arx(v.co[:]),
                blender_pos_to_arx(normal)
            ))
        
        originVertexIndex = -1
        for index, vert in enumerate(verts):
            if vert.xyz == (0.0, 0.0, 0.0):
                #self.log.info("Origin vertex found at index {}".format(index))
                originVertexIndex = index
        
        if originVertexIndex == -1:
            self.log.info("Origin vertex not found, adding new one")
            originVertexIndex = len(verts)
            verts.append(FtlVertex((0.0, 0.0, 0.0), (0.0, 0.0, 0.0)))
        
        metadata = FtlMetadata(name=obj.get('arx.ftl.name', ''), org=originVertexIndex)
        
        # Add action point vertices
        actions = []
        for o in bpy.data.objects:
            if o.type == 'EMPTY' and o.parent == obj:
                nameParts = o.name.split(".")
                name = nameParts[0]
                if o.parent_type == 'VERTEX':
                    actionVertexIndex = o.parent_vertices[0]
                    actions.append(FtlAction(name, actionVertexIndex))
                elif o.parent_type == 'OBJECT':
                    actionVertexIndex = len(verts)
                    verts.append(FtlVertex((o.location[0], o.location[1], o.location[2]), (0.0, 0.0, 0.0)))
                    actions.append(FtlAction(name, actionVertexIndex))
                else:
                    self.log.warn("Unhandled empty parent type {}".format(o.parent_type))

        uvData = bm.loops.layers.uv.verify()

        allXYZ = [v.xyz for v in verts]
        faces = []

        for f in bm.faces:
            vertexIndices = []
            vertexUvs = []
            vertexNormals = []
            for c in f.loops:
                vertexIndices.append(allXYZ.index(blender_pos_to_arx(c.vert.co[:])))
                vertexUvs.append((c[uvData].uv[0], 1 - c[uvData].uv[1]))
                vertexNormals.append((c.vert.normal[0], c.vert.normal[1], c.vert.normal[2]))
            mat = f.material_index

            faceType = f[arxFaceType]
            transval = f[arxTransVal]

            faces.append(FtlFace(
                vids=tuple(vertexIndices),
                uvs=vertexUvs,
                texid=mat,
                facetype=faceType,
                transval=transval,
                normal=vertexNormals
            ))

        grps = []
        sels = []

        dvert_lay = bm.verts.layers.deform.active

        for grp in obj.vertex_groups:
            v = []
            if dvert_lay is not None:
                for vert in bm.verts:
                    dvert = vert[dvert_lay]
                    if grp.index in dvert:
                        v.append(vert.index)

            s = grp.name.split(":")
            if s[0] == "grp":
                armature = bpy.data.armatures.get("/".join(canonicalId))
                if armature:
                    origin = armature.bones[grp.name]["OriginVertex"]
                else:
                    # TODO hardcoded use of mesh vertex at index 0
                    origin = 0
                
                # TODO get the real data from armature
                parentIndex = -1
                grps.append(FtlGroup(s[2], origin, v, parentIndex))
            else:
                sels.append(FtlSelection(s[2], v))

        matNames = []
        for o in bpy.data.objects:
            for ms in o.material_slots:
                mat = ms.material
                # strip -mat suffix and decorate with bullshit
                matNames.append('GRAPH\\OBJ3D\\TEXTURES\\' + mat.name[:-4] + ".FOO")

        bm.free()

        return FtlData(
            metadata=metadata,
            verts=verts,
            faces=faces,
            mats=matNames,
            groups=grps,
            actions=actions,
            sels=sels
        )

    def saveFile(self, path):

        data = self.toFtlData()

        binData = self.ftlSerializer.write(data)

        with open(path, 'wb') as f:
            f.write(binData)

        self.log.debug("Written %i bytes to file %s" % (len(binData), path))


# Copyright 2014-2020 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.

from ctypes import (
    LittleEndianStructure,
    c_char,
    c_uint32,
    c_int16,
    c_int32,
    c_float
)

from .dataCommon import SavedVec3, PolyTypeFlag

class UNIQUE_HEADER(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("path",             c_char * 256),
        ("count",            c_int32),
        ("version",          c_float),
        ("uncompressedsize", c_int32),
        ("pad",              c_int32 * 3)
    ]

class UNIQUE_HEADER3(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("path",  c_char * 256), # In the c code this is currently in a separate struct
        ("check", c_char * 512)
    ]

class FAST_SCENE_HEADER(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("version",     c_float),
        ("sizex",       c_int32),
        ("sizez",       c_int32),
        ("nb_textures", c_int32),
        ("nb_polys",    c_int32),
        ("nb_anchors",  c_int32),
        ("playerpos",   SavedVec3),
        ("Mscenepos",   SavedVec3),
        ("nb_portals",  c_int32),
        ("nb_rooms",    c_int32)
    ]

class FAST_TEXTURE_CONTAINER(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("tc",   c_int32),
        ("temp", c_int32),
        ("fic",  c_char * 256)
    ]

class FAST_SCENE_INFO(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("nbpoly",     c_int32),
        ("nbianchors", c_int32),
    ]

class FAST_VERTEX(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("sy",  c_float),
        ("ssx", c_float),
        ("ssz", c_float),
        ("stu", c_float),
        ("stv", c_float)
    ]

class FAST_EERIEPOLY(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("v",        FAST_VERTEX * 4),
        ("tex",      c_int32),
        ("norm",     SavedVec3),
        ("norm2",    SavedVec3),
        ("nrml",     SavedVec3 * 4),
        ("transval", c_float),
        ("area",     c_float),
        ("type",     PolyTypeFlag),
        ("room",     c_int16),
        ("paddy",    c_int16)
    ]

class FAST_ANCHOR_DATA(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("pos",       SavedVec3),
        ("radius",    c_float),
        ("height",    c_float),
        ("nb_linked", c_int16),
        ("flags",     c_int16)
    ]

class SavedTextureVertex(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("pos",      SavedVec3),
        ("rhw",      c_float),
        ("color",    c_uint32),
        ("specular", c_uint32),
        ("tu",       c_float),
        ("tv",       c_float)
    ]

class SAVE_EERIEPOLY(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("type",     c_int32),
        ("min",      SavedVec3),
        ("max",      SavedVec3),
        ("norm",     SavedVec3),
        ("norm2",    SavedVec3),
        ("v",        SavedTextureVertex * 4),
        ("tv",       SavedTextureVertex * 4),
        ("nrml",     SavedVec3 * 4),
        ("tex",      c_int32),
        ("center",   SavedVec3),
        ("transval", c_float),
        ("area",     c_float),
        ("room",     c_int16),
        ("misc",     c_int16)
    ]

class EERIE_SAVE_PORTALS(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("poly",      SAVE_EERIEPOLY),
        ("room_1",    c_int32),
        ("room_2",    c_int32),
        ("useportal", c_int16),
        ("paddy",     c_int16)
    ]
    
class EERIE_SAVE_ROOM_DATA(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("nb_portals", c_int32),
        ("nb_polys",   c_int32),
        ("padd",       c_int32 * 6)
    ]

class FAST_EP_DATA(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("px",   c_int16),
        ("py",   c_int16),
        ("idx",  c_int16),
        ("padd", c_int16)
    ]
    
class ROOM_DIST_DATA_SAVE(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ("distance", c_float),
        ("startpos", SavedVec3),
        ("endpos",   SavedVec3),
    ]


from collections import namedtuple

FtsData = namedtuple('FtsData', ['sceneOffset', 'textures', 'cells', 'anchors', 'portals'])

import logging

from ctypes import sizeof
from .lib import ArxIO

class FtsSerializer(object):
    def __init__(self, ioLib):
        self.log = logging.getLogger('FtsSerializer')
        self.ioLib = ioLib

    def read_fts(self, data) -> FtsData:
        """If you want to read a fts file use read_fts_container"""

        pos = 0
        ftsHeader = FAST_SCENE_HEADER.from_buffer_copy(data, pos)
        pos += sizeof(FAST_SCENE_HEADER)
        self.log.debug("Fts Header version: %f" % ftsHeader.version)
        self.log.debug("Fts Header size x,z: %i,%i" % (ftsHeader.sizex, ftsHeader.sizez))
        self.log.debug("Fts Header playerpos: %f,%f,%f" % (ftsHeader.playerpos.x, ftsHeader.playerpos.y, ftsHeader.playerpos.z))
        self.log.debug("Fts Header Mscenepos: %f,%f,%f" % (ftsHeader.Mscenepos.x, ftsHeader.Mscenepos.y, ftsHeader.Mscenepos.z))
        sceneOffset = (ftsHeader.Mscenepos.x, ftsHeader.Mscenepos.y, ftsHeader.Mscenepos.z)

        texturesType = FAST_TEXTURE_CONTAINER * ftsHeader.nb_textures
        textures = texturesType.from_buffer_copy(data, pos)
        pos += sizeof(texturesType)
        self.log.debug("Loaded %i textures" % len(textures))

        #for i in textures:
        #    log.info(i.fic.decode('iso-8859-1'))
        
        cells = [[None for x in range(ftsHeader.sizex)] for x in range(ftsHeader.sizez)]
        for z in range(ftsHeader.sizez):
            for x in range(ftsHeader.sizex):
                cellHeader = FAST_SCENE_INFO.from_buffer_copy(data, pos)
                pos += sizeof(FAST_SCENE_INFO)

                try:
                    if cellHeader.nbpoly <= 0:
                        cells[z][x] = None
                    else:
                        polysType = FAST_EERIEPOLY * cellHeader.nbpoly
                        polys = polysType.from_buffer_copy(data, pos)
                        pos += sizeof(polysType)

                        cells[z][x] = polys
                except ValueError as e:
                    print("Failed reading cell data, x:%i z:%i polys:%i" % (x, z, cellHeader.nbpoly))
                    raise e

                
                if cellHeader.nbianchors > 0:
                    AnchorsArrayType = c_int32 * cellHeader.nbianchors
                    anchors = AnchorsArrayType.from_buffer_copy(data, pos)
                    pos += sizeof(AnchorsArrayType)
                    
        anchors = []
        for i in range(ftsHeader.nb_anchors):
            anchor = FAST_ANCHOR_DATA.from_buffer_copy(data, pos)
            pos += sizeof(FAST_ANCHOR_DATA)
            
            if anchor.nb_linked > 0:
                LinkedAnchorsArrayType = c_int32 * anchor.nb_linked
                linked = LinkedAnchorsArrayType.from_buffer_copy(data, pos)
                pos += sizeof(LinkedAnchorsArrayType)
                anchors.append( ((anchor.pos.x, anchor.pos.y, anchor.pos.z), linked) )
            else:
                anchors.append( ((anchor.pos.x, anchor.pos.y, anchor.pos.z), []) )
        
        portals = []
        for i in range(ftsHeader.nb_portals):
            portal = EERIE_SAVE_PORTALS.from_buffer_copy(data, pos)
            pos += sizeof(EERIE_SAVE_PORTALS)
            portals.append(portal)

        for i in range(ftsHeader.nb_rooms + 1): # Off by one in data
            room = EERIE_SAVE_ROOM_DATA.from_buffer_copy(data, pos)
            pos += sizeof(EERIE_SAVE_ROOM_DATA)
            
            if room.nb_portals > 0:
                PortalsArrayType = c_int32 * room.nb_portals
                portals2 = PortalsArrayType.from_buffer_copy(data, pos)
                pos += sizeof(PortalsArrayType)
                
            if room.nb_polys > 0:
                PolysArrayType = FAST_EP_DATA * room.nb_polys
                polys2 = PolysArrayType.from_buffer_copy(data, pos)
                pos += sizeof(PolysArrayType)
        
        for i in range(ftsHeader.nb_rooms):
            for j in range(ftsHeader.nb_rooms):
                dist = ROOM_DIST_DATA_SAVE.from_buffer_copy(data, pos)
                pos += sizeof(ROOM_DIST_DATA_SAVE)
                
        self.log.debug("Loaded %i bytes of %i" % (pos, len(data)))

        return FtsData(
            sceneOffset=sceneOffset,
            textures=textures,
            cells=cells,
            anchors=anchors,
            portals=portals
        )

    def read_fts_container(self, filepath) -> FtsData:
        f = open(filepath, "rb")
        data = f.read()
        f.close()

        self.log.debug("Loaded %i bytes from file %s" % (len(data), filepath))
        
        pos = 0
        
        primaryHeader = UNIQUE_HEADER.from_buffer_copy(data, pos)
        pos += sizeof(UNIQUE_HEADER)
        self.log.debug("Header path: %s" % primaryHeader.path.decode('iso-8859-1'))
        self.log.debug("Header count: %i" % primaryHeader.count)
        self.log.debug("Header version: %f" % primaryHeader.version)
        self.log.debug("Header uncompressedsize: %i" % primaryHeader.uncompressedsize)
            
        secondaryHeadersType = UNIQUE_HEADER3 * primaryHeader.count
        secondaryHeaders = secondaryHeadersType.from_buffer_copy(data, pos)
        pos += sizeof(secondaryHeadersType)
        
        for h in secondaryHeaders:
            self.log.debug("Header2 path: %s" % h.path.decode('iso-8859-1'))
        
        uncompressed = self.ioLib.unpack(data[pos:])
        
        if primaryHeader.uncompressedsize != len(uncompressed):
            self.log.warn("Uncompressed size mismatch, expected %i actual %i" % (primaryHeader.uncompressedsize, len(uncompressed)))
        
        return self.read_fts(uncompressed)



# Copyright 2019-2020 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.


import os
import logging

import bpy
from mathutils import Vector

log = logging.getLogger('Materials')

def arx_create_image(rootDirectory, relativePath):
    extensions = [".png", ".jpg", ".jpeg", ".bmp", ".tga"]
    for ext in extensions:
        fullPath = os.path.join(rootDirectory, relativePath) + ext
        if os.path.exists(fullPath):
            break

    if not os.path.exists(fullPath):
        log.warning("Texture not found: %s" % relativePath)
        return None

    if relativePath in bpy.data.images:
        img = bpy.data.images[relativePath]
    else:
        img = bpy.data.images.load(fullPath)
        img.name = relativePath

    return img

def arx_create_material_nodes(material, image):
    tree = material.node_tree

    n_output = None
    n_principled = None
    for node in tree.nodes:
        if node.bl_idname == 'ShaderNodeOutputMaterial':
            n_output = node
            continue
        if node.bl_idname == 'ShaderNodeBsdfPrincipled':
            n_principled = node
            continue

    # Remove default principled bsdf node
    tree.nodes.remove(n_principled)

    n_output.location = (0, 0)

    n_arx = tree.nodes.new('ShaderNodeGroup')
    n_arx.name = 'n_arx'
    n_arx.location = (-200, 0)
    n_arx.node_tree = arx_get_material_node_group()

    n_tex_image = tree.nodes.new('ShaderNodeTexImage')
    n_tex_image.name = 'n_tex_image'
    n_tex_image.location = (-600, 0)

    tree.links.new(n_arx.outputs['Shader'], n_output.inputs['Surface'])
    tree.links.new(n_tex_image.outputs['Color'], n_arx.inputs['Color'])

    n_tex_image.image = image

def createMaterial(rootDirectory, textureName) -> bpy.types.Material:

    relativePath, fileExtension = os.path.splitext(textureName.replace("\\", "/").lower())
    foo, fileName = os.path.split(relativePath)
    
    matName = fileName + "-mat"
    
    mats = bpy.data.materials
    matches = [m for m in mats if m.name == matName]
    if matches:
        return matches[0]

    image = arx_create_image(rootDirectory, relativePath)

    mat = bpy.data.materials.new(matName)
    mat.use_nodes = True
    mat.blend_method = 'HASHED'
    arx_create_material_nodes(mat, image)

    return mat

    if relativePath in bpy.data.textures:
        tex = bpy.data.textures[relativePath]
    else:
        tex = bpy.data.textures.new(relativePath, type='IMAGE')
        tex.image = img
    
    mtex = mat.texture_slots.add()
    mtex.texture = tex
    mtex.texture_coords = 'UV'
    mtex.use_map_color_diffuse = True

    # This is a hack for testing !
    if "[stone]_fire_torch01" in relativePath:
        print("=================_fire_torch01")
        mat.use_transparency = True
        mat.transparency_method = "Z_TRANSPARENCY"
        mat.offset_z = 1
        img.use_alpha = False

        mtex.use_map_alpha = True
        mtex.alpha_factor = -1
        mtex.blend_type = 'SUBTRACT'

        black = getBlackTexture()
        slot = mat.texture_slots.add()
        slot.texture = black

    return mat

#from arx_addon.materials import arx_create_material_node_group
#arx_create_material_node_group()
#am = D.node_groups['Arx Material']


def arx_get_material_node_group():
    arx_material_node_group_name = 'Arx Material'

    if arx_material_node_group_name in bpy.data.node_groups:
        return bpy.data.node_groups[arx_material_node_group_name]

    group = bpy.data.node_groups.new(arx_material_node_group_name, 'ShaderNodeTree')

    group.outputs.new('NodeSocketShader', 'Shader')
    group.inputs.new('NodeSocketColor', 'Color')

    n_out = group.nodes.new('NodeGroupOutput')
    n_out.name = 'n_out'
    n_out.location = Vector((0.0, 0.0))

    n_mix = group.nodes.new('ShaderNodeMixShader')
    n_mix.name = 'n_mix'
    n_mix.location = Vector((-220.0, 0.0))

    n_transparent = group.nodes.new('ShaderNodeBsdfTransparent')
    n_transparent.name = 'n_transparent'
    n_transparent.location = Vector((-520.0, -140.0))

    n_diffuse = group.nodes.new('ShaderNodeBsdfDiffuse')
    n_diffuse.name = 'n_diffuse'
    n_diffuse.location = Vector((-520.0, 0.0))

    n_math1 = group.nodes.new('ShaderNodeMath')
    n_math1.name = 'n_math1'
    n_math1.location = Vector((-520.0, 180.0))
    n_math1.operation = 'FLOOR'

    n_geometry = group.nodes.new('ShaderNodeNewGeometry')
    n_geometry.name = 'n_geometry'
    n_geometry.location = Vector((-760.0, 420.0))

    n_colMix = group.nodes.new('ShaderNodeMixRGB')
    n_colMix.name = 'n_colMix'
    n_colMix.blend_type = 'MULTIPLY'
    n_colMix.inputs['Fac'].default_value = 1.0
    n_colMix.location = Vector((-730.0, -238.0))

    n_vertexColor = group.nodes.new('ShaderNodeVertexColor')
    n_vertexColor.name = 'n_vertexColor'
    n_vertexColor.layer_name = 'light-color'
    n_vertexColor.location = Vector((-970.0, -91.0))

    n_in = group.nodes.new('NodeGroupInput')
    n_in.name = 'n_in'
    n_in.location = Vector((-1300.0, 0.0))

    group.links.new(n_mix.outputs['Shader'], n_out.inputs['Shader'])

    group.links.new(n_math1.outputs['Value'], n_mix.inputs['Fac'])
    group.links.new(n_diffuse.outputs['BSDF'], n_mix.inputs[1])
    group.links.new(n_transparent.outputs['BSDF'], n_mix.inputs[2])

    group.links.new(n_geometry.outputs['Backfacing'], n_math1.inputs['Value'])

    group.links.new(n_colMix.outputs['Color'], n_diffuse.inputs['Color'])
    group.links.new(n_in.outputs['Color'], n_colMix.inputs['Color1'])
    group.links.new(n_vertexColor.outputs['Color'], n_colMix.inputs['Color2'])

    return group



# Copyright 2015-2020 Arx Libertatis Team (see the AUTHORS file)
#
# This file is part of Arx Libertatis.
#
# Arx Libertatis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Arx Libertatis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Arx Libertatis. If not, see <http://www.gnu.org/licenses/>.

import os
import io
import re
from collections import namedtuple, OrderedDict


def splitPath(path):
    result = []
    while 1:
        path, folder = os.path.split(path)
        if folder != "":
            result.append(folder)
        else:
            if path != "":
                result.append(path)
            break
    result.reverse()
    return result;


EntityInstance = namedtuple("ArxInstance", ["id", "script"])
EntityData = namedtuple("EntityData", ["path", "script", "icon", "instances"])


class Entities:
    def __init__(self):
        self.paths = ["graph/obj3d/interactive"]
        self.danglingPaths = []
        self.data = {}

    def update(self, absroot):

        for root, dirs, files in os.walk(absroot):
            split = splitPath(root[len(absroot) + len(self.paths):])

            mainScript = None

            for f in files:
                name, ext = os.path.splitext(f)
                if ext == ".asl" and name == split[-1]:
                    mainScript = f

            if mainScript is not None:
                files.remove(mainScript)

                icon = None

                for f in files:
                    fullPath = os.path.join(root, f)
                    name, ext = os.path.splitext(f)
                    if name.endswith("[icon]"):
                        icon = fullPath
                    else:
                        self.danglingPaths.append(fullPath)

                instances = []

                for d in dirs:
                    if d.startswith(split[-1]):
                        instanceId = int(d[-4:])

                        instanceDir = os.path.join(root, d)
                        instanceFiles = os.listdir(instanceDir)

                        instanceScript = None

                        for f in instanceFiles:
                            if os.path.isfile(os.path.join(instanceDir, f)):
                                if f == mainScript:
                                    instanceScript = os.path.join(root, d, f)
                                else:
                                    self.danglingPaths.append(os.path.join(root, d, f))
                            else:
                                self.danglingPaths.append(os.path.join(root, d, f))

                        instances.append(EntityInstance(instanceId, instanceScript))
                    else:
                        self.danglingPaths.append(os.path.join(root, d))

                instances.sort(key=lambda x: x.id)

                key = tuple(split)
                if key in self.data:
                    print("Warn: Duplicated name !!")

                e = EntityData(root, mainScript, icon, instances)
                self.data[key] = e


    def parseResourceReferences(self):
        for key in self.data:
            value = self.data[key]

            scriptPath = os.path.join(value.path, value.script)

            with io.open(scriptPath, 'r', encoding='latin1') as f:
                lines = f.readlines()
                for line in lines:
                    if "loadanim" in line.lower():
                        print(line)


ModelData = namedtuple("ModelData", ["path", "model", "tweaks"])

class Models:
    def __init__(self):
        self.paths = ["game/graph/obj3d/interactive"]
        self.danglingPaths = []
        self.data = {}

    def update(self, absroot):

        for root, dirs, files in os.walk(absroot):
            split = splitPath(root[len(absroot) + len(self.paths):])

            model = None

            for f in files:
                name, ext = os.path.splitext(f)
                if ext == ".ftl":
                    if name == split[-1]:
                        model = f
                    else:
                        self.danglingPaths.append(os.path.join(root, f))

            if model is not None:
                files.remove(model)

                for f in files:
                    if f.endswith(".unpack"):
                        pass
                    else:
                        self.danglingPaths.append(os.path.join(root, f))

                tweaks = []

                for d in dirs:
                    if d == "tweaks":
                        tweakDir = os.path.join(root, d)
                        tweakFiles = os.listdir(tweakDir)
                        for f in tweakFiles:
                            if f.endswith(".ftl"):
                                tweaks.append(f)
                            elif f.endswith(".unpack"):
                                pass
                            else:
                                self.danglingPaths.append(os.path.join(root, d, f))
                    else:
                        self.danglingPaths.append(os.path.join(root, d))

                dirs.clear()

                key = tuple(split)
                if key in self.data:
                    print("Warn: Duplicated name !!")

                m = ModelData(root, model, tweaks)
                self.data[key] = m


class Speeches:
    def __init__(self):
        self.paths = ["speech"]
        self.danglingPaths = []
        self.files = {}

    def update(self, absroot):
        dirs = os.listdir(absroot)
        for langName in dirs:
            l = os.path.join(absroot, langName)
            if os.path.isdir(l):
                foo = []
                sounds = os.listdir(l)
                for sound in sounds:
                    f = os.path.join(l, sound)
                    name, ext = os.path.splitext(sound)
                    if not os.path.isdir(f) and ext == ".wav":
                        foo.append(sound)
                    else:
                        self.danglingPaths.append(f)
                self.files[langName] = foo
            else:
                self.danglingPaths.append(l)


class LevelInfos(object):
    def __init__(self):
        self.fts = None
        self.dlf = None
        self.llf = None
        self.map = None
        self.load = None


class Levels:
    def __init__(self):
        self.paths = ["graph/levels", "game/graph/levels"]
        self.danglingPaths = []
        self.levels = OrderedDict()

    def update(self, absroot):
        dirs = os.listdir(absroot)
        for lvlName in dirs:
            l = os.path.join(absroot, lvlName)
            if not os.path.isdir(l):
                self.danglingPaths.append(l)
                continue
            if not lvlName.startswith('level'):
                self.danglingPaths.append(l)
                continue

            area_index = int(lvlName[5:])

            info = self.levels.setdefault(area_index, LevelInfos())
            
            foo = os.listdir(l)
            for bar in foo:
                blubb = os.path.join(l, bar)
                if not os.path.isdir(blubb):
                    name, ext = os.path.splitext(bar)
                    if ext == ".fts":
                        info.fts = blubb
                    elif ext == ".llf":
                        info.llf = blubb
                    elif ext == ".dlf":
                        info.dlf = blubb
                    elif bar == "map.bmp":
                        info.map = blubb
                    elif name == "loading":
                        info.load = blubb
                    else:
                        self.danglingPaths.append(blubb)
                else:
                    self.danglingPaths.append(blubb)
        self.levels = OrderedDict(sorted(self.levels.items()))


class Cinematics:
    def __init__(self):
        self.paths = ["graph/interface"] # TODO why not "graph/interface/illustrations"
        self.danglingPaths = []
        self.cins = []
        self.textures = {}

    def update(self, root):
        root = os.path.join(root, 'illustrations') # TODO hack
        sub = os.listdir(root)
        for s in sub:
            foo = os.path.join(root, s)
            if not os.path.isdir(foo):
                name, ext = os.path.splitext(s)
                if ext == ".cin":
                    self.cins.append(foo)
                else:
                    self.danglingPaths.append(foo)
            else:
                if s == "illust":
                    bar = os.listdir(foo)
                    for b in bar:
                        name, ext = os.path.splitext(b)
                        if ext == ".tga":
                            self.textures[name] = b
                        else:
                            self.danglingPaths.append(b)
                else:
                    self.danglingPaths.append(foo)


class Textures:
    def __init__(self):
        self.paths = ["graph/particles", "graph/interface", "graph/obj3d/textures"]
        self.danglingPaths = []
        self.textures = []

    def update(self, root):
        pass


class Animations:
    def __init__(self):
        self.paths = ["graph/obj3d/anims"]
        self.danglingPaths = []
        self.amins = []
        self.data = {}

    def update(self, root):
        for root, dirs, files in os.walk(root):

            for f in files:
                foo = os.path.join(root, f)
                name, ext = os.path.splitext(f)
                if ext == ".tea":
                    self.amins.append(foo)
                    self.data[name] = foo
                else:
                    self.danglingPaths.append(foo)


class AudioEffects:
    def __init__(self):
        self.paths = ["sfx"]
        self.danglingPaths = []
        self.effects = []
        self.ambiances = []
        self.environments = []

    def update(self, root):
        for root, dirs, files in os.walk(root):

            for f in files:
                foo = os.path.join(root, f)
                name, ext = os.path.splitext(f)
                if ext == ".wav":
                    self.effects.append(name)
                elif ext == ".amb":
                    self.ambiances.append(name)
                elif ext == ".aef":
                    self.environments.append(name)
                else:
                    self.danglingPaths.append(foo)

class AudioSpeech:
    def __init__(self):
        self.paths = ["speech"]
        self.danglingPaths = []
        self.speeches = {}

    def update(self, root):
        langDirNames = os.listdir(root)
        for langDirName in langDirNames:
            langDir = os.path.join(root, langDirName)
            fileNames = os.listdir(langDir)
            files = []
            for fileName in fileNames:
                files.append(fileName)

            self.speeches[langDirName] = files


class ArxFiles(object):
    def __init__(self, rootPath):
        self.rootPath = str(rootPath)
        self.allFiles = set()

        self.entities = Entities()
        self.models = Models()
        self.speeches = Speeches()
        self.levels = Levels()
        self.cinematics = Cinematics()
        self.animations = Animations()
        self.textures = Textures()
        self.audioEffects = AudioEffects()
        self.audioSpeech = AudioSpeech()

        self.handlers = [self.entities, self.models, self.speeches, self.levels, self.cinematics, self.animations,
                         self.textures, self.audioEffects, self.audioSpeech]

        self.danglingPaths = []

    def updateAll(self):
        for root, dirs, files in os.walk(self.rootPath):
            relRoot = os.path.relpath(root, self.rootPath)

            for f in files:
                relFile = os.path.join(relRoot, f)
                self.allFiles.add(relFile)

        for root, dirs, files in os.walk(self.rootPath):
            relRoot = os.path.relpath(root, self.rootPath)

            for h in self.handlers:
                if relRoot in h.paths:
                    h.update(root)
                    dirs.clear()
                    files.clear()
                    continue

            for f in files:
                self.danglingPaths.append(os.path.join(root, f))

                # self.entities.parseResourceReferences()
